"""This module provides the FourierDomainExpression class to represent
f-domain (Fourier domain) expressions.

Copyright 2014--2023 Michael Hayes, UCECE

"""

from __future__ import division
from .domains import FourierDomain
from .expr import Expr, expr, expr_make
from .functions import integrate
from .inverse_dtft import IDTFT
from .inverse_fourier import inverse_fourier_transform
from .state import state, validate
from .sym import fsym, ssym, tsym, omegasym, pi
from .sym import nsym, dt, oo
from .units import u as uu
from .utils import factor_const
from sympy import Integral, Expr as symExpr

__all__ = ('FourierDomainExpression', 'fexpr')


class FourierDomainExpression(FourierDomain, Expr):

    """Fourier domain expression or symbol."""

    var = fsym

    def __init__(self, val, **assumptions):

        check = assumptions.pop('check', True)
        assumptions['real'] = True
        super(FourierDomainExpression, self).__init__(val, **assumptions)

        expr = self.expr

        if check and not expr.has(fsym):
            if expr.has(ssym):
                validate(state.s_in_f,
                         'f-domain expression % s depends on s' % expr)
            if expr.has(tsym):
                validate(state.t_in_f,
                         'f-domain expression %s depends on t' % expr)
            if expr.has(omegasym):
                validate(state.w_in_f,
                         'f-domain expression %s depends on omega' % expr)

    def as_expr(self):
        return FourierDomainExpression(self)

    @property
    def energy(self):
        """Return signal energy."""

        E = integrate(abs(self)**2, (f, -oo, oo))

        # FIXME: result should not be a Fourier-domain expression
        return E

    def inverse_fourier(self, evaluate=True, **assumptions):
        """Attempt inverse Fourier transform."""

        result = inverse_fourier_transform(
            self.expr, self.var, tsym, evaluate=evaluate)

        return self.change(result, 'time', units_scale=uu.Hz, **assumptions)

    def IFT(self, evaluate=True, **assumptions):
        """Convert to time domain.  This is an alias for inverse_fourier."""

        return self.inverse_fourier(evaluate=evaluate, **assumptions)

    def IDTFT(self, evaluate=True, **assumptions):
        """Convert to discrete-time domain using inverse discrete-time
        Fourier transform."""

        result = IDTFT(self.expr, self.var, nsym, evaluate=evaluate)

        return self.change(result, 'discrete time', units_scale=uu.Hz,
                           **assumptions)

    def time(self, **assumptions):
        return self.inverse_fourier(**assumptions)

    def norm_fourier(self, **assumptions):
        """Convert to normalized Fourier domain."""
        from .symbols import F
        from .sym import dt

        result = self.subs(F / dt)
        return result

    def angular_fourier(self, **assumptions):
        """Convert to angular Fourier domain."""
        from .symbols import omega

        result = self.subs(omega / (2 * pi))
        return result

    def norm_angular_fourier(self, **assumptions):
        """Convert to normalized angular Fourier domain."""
        from .symbols import Omega
        from .sym import dt

        result = self.subs(Omega / (2 * pi * dt))
        return result

    def frequency_response(self, **assumptions):
        """Convert to frequency response domain."""

        return self.laplace(**assumptions).frequency_response()

    def angular_frequency_response(self, **assumptions):
        """Convert to angular frequency response domain."""

        return self.laplace(**assumptions).angular_frequency_response()

    def laplace(self, **assumptions):
        """Determine one-sided Laplace transform with 0- as the lower limit."""

        result = self.time(**assumptions).laplace()
        return result

    def phasor(self, **assumptions):
        """Convert to phasor domain."""

        return self.time(**assumptions).phasor(**assumptions)

    def plot(self, fvector=None, plot_type=None, **kwargs):
        """Plot frequency response at values specified by `fvector`.

        If `fvector` is a tuple, this sets the frequency limits.

        `plot_type` - 'dB-phase', 'dB-phase-degrees', 'mag-phase',
        'mag-phase-degrees', 'real-imag', 'mag', 'phase',
        'phase-degrees', 'real', or 'imag'.

        The default `plot_type` for complex data is `dB-phase`.

        `kwargs` include:
        `axes` - the plot axes to use otherwise a new figure is created
        `xlabel` - the x-axis label
        `ylabel` - the y-axis label
        `ylabel2` - the second y-axis label if needed, say for mag and phase
        `xscale` - the x-axis scaling, say for plotting as ms
        `yscale` - the y-axis scaling, say for plotting mV
        `norm` - use normalized frequency
        `dbmin` - the smallest value to plot in dB (default -120)
        `plot_deltas` - plot Dirac deltas as arrows
        in addition to those supported by the matplotlib plot command.

        The plot axes are returned.  This is a tuple for magnitude/phase or
        real/imaginary plots.

        There are many plotting options, see lcapy.plot and
        matplotlib.pyplot.plot.

        For example:
            V.plot(fvector, log_frequency=True)
            V.real.plot(fvector, color='black')
            V.phase.plot(fvector, color='black', linestyle='--')

        By default complex data is plotted as separate plots of
        magnitude (dB) and phase.

        """

        from .plot import plot_frequency
        return plot_frequency(self, fvector, plot_type=plot_type, **kwargs)

    def bode_plot(self, fvector=None, unwrap=True, **kwargs):
        """Plot frequency response for a frequency-domain phasor as a Bode
        plot (but without the straight line approximations).  fvector
        specifies the frequencies.  If it is a tuple (f1, f2), it sets
        the frequency limits.   Since a logarithmic frequency scale is used,
        f1 must be greater than 0.

        `unwrap` controls phase unwrapping (default True).

        For more info, see `plot`.
        """

        from .plot import plot_bode
        return plot_bode(self, fvector, unwrap=unwrap, **kwargs)

    def nyquist_plot(self, fvector=None, log_frequency=True, **kwargs):
        """Plot frequency response as a Nyquist plot (imaginary part versus
        real part).  fvector specifies the frequencies.  If it is
        a tuple (f1, f2), it sets the frequency limits as (f1, f2).

        `npoints` set the number of plotted points.

        The unit circle is shown by default.  This can be disabled with `unitcircle=False`.
        """

        from .plot import plot_nyquist
        return plot_nyquist(self, fvector, log_frequency=log_frequency, **kwargs)

    def nichols_plot(self, fvector=None, log_frequency=True, **kwargs):
        """Plot frequency response as a Nichols plot (dB part versus
        phase).  fvector specifies the frequencies.  If it is
        a tuple (f1, f2), it sets the frequency limits as (f1, f2).

        `npoints` set the number of plotted points.
        """

        from .plot import plot_nichols
        return plot_nichols(self, fvector, log_frequency=log_frequency, **kwargs)


def fexpr(arg, **assumptions):
    """Create FourierDomainExpression object.  If `arg` is fsym return f"""

    if arg is fsym:
        return f
    return expr_make('fourier', arg, **assumptions)


f = FourierDomainExpression('f')
f.units = uu.Hz



from lcapy import *
from lcapy.sym import tausym
from lcapy.texpr import TimeDomainExpression
from lcapy.omegaexpr import AngularFourierDomainExpression
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_functions(self):

        self.assertEqual(sinc(0), 1, "sinc(0)")
        self.assertEqual(sinc(1), 0, "sinc(1)")
        self.assertEqual(rect(0), 1, "rect(0)")
        self.assertEqual(rect(1), 0, "rect(1)")
        self.assertEqual(ui(1), 0, "ui(1)")
        self.assertEqual(ui(0), 1, "ui(0)")
        self.assertEqual(us(1), 1, "us(1)")
        self.assertEqual(us(0), 1, "us(0)")
        self.assertEqual(us(-1), 0, "us(-1)")

        self.assertEqual(rect(n).rewrite(), us(
            n + 1 / 2) - us(n - 1 / 2), "rect(n)")

        r = Integral(u(t - tausym), (tausym, 0, t))
        e = u(t).convolve(u(t)).simplify()

        self.assertEqual(e, r, "simplify integral with u(t)")

        f0 = symbol('f_0')
        c = cos(2 * pi * f0 * t)
        self.assertEqual(type(c), TimeDomainExpression, "cos")

        a = atan2(1, omega)
        self.assertEqual(type(a), AngularFourierDomainExpression, "atan2")

        self.assertEqual(sign(-n).evaluate(0), 1, "sign(-n)")
        self.assertEqual(sign(-n).evaluate(1), -1, "sign(-n)")

    def test_heaviside(self):

        self.assertEqual((u(t) * 3 * u(t)).simplify(),
                         3 * u(t), "u(t) * 3 * u(t)")

    def test_dirac_delta_simplify(self):

        self.assertEqual(expr('delta(t) * exp(-3 * t)').simplify(),
                         expr('delta(t)'),
                         "delta(t) * exp(-3 * t)")

        self.assertEqual(expr('exp(-3 * t) * delta(t)').simplify(),
                         expr('delta(t)'),
                         "exp(-3 * t) * delta(t)")

        self.assertEqual((rect(t / 4) * delta(t - 1)).simplify(),
                         delta(t - 1),
                         "rect(t / 4) * delta(t - 1)")

    def test_sinc(self):

        a = sinc(n).evaluate(2)
        self.assertEqual(abs(a) < 1e-16, True, "sinc evaluate incorrect.")

        a = sincn(n).evaluate(2)
        self.assertEqual(abs(a) < 1e-16, True, "sincn evaluate incorrect.")

        a = expr('sinc(n)').evaluate(2)
        self.assertEqual(abs(a) < 1e-16, True, "'sinc' evaluate incorrect.")

        a = expr('sincn(n)').evaluate(2)
        self.assertEqual(abs(a) < 1e-16, True, "'sincn' evaluate incorrect.")




# This file is automatically generated by makeclasses.py

from .units import units, u as uu
from .voltagemixin import VoltageMixin
from .currentmixin import CurrentMixin
from .admittancemixin import AdmittanceMixin
from .impedancemixin import ImpedanceMixin
from .transfermixin import TransferMixin
from .voltagesquaredmixin import VoltageSquaredMixin
from .currentsquaredmixin import CurrentSquaredMixin
from .admittancesquaredmixin import AdmittanceSquaredMixin
from .impedancesquaredmixin import ImpedanceSquaredMixin
from .powermixin import PowerMixin
from sympy import sqrt, S
from .cexpr import ConstantDomainExpression
from .cexpr import ConstantTimeDomainExpression
from .cexpr import ConstantFrequencyResponseDomainExpression
from .fexpr import FourierDomainExpression
from .jfexpr import FrequencyResponseDomainExpression
from .jomegaexpr import AngularFrequencyResponseDomainExpression
from .kexpr import DiscreteFourierDomainExpression
from .nexpr import DiscreteTimeDomainExpression
from .noisefexpr import FourierNoiseDomainExpression
from .noiseomegaexpr import AngularFourierNoiseDomainExpression
from .normfexpr import NormFourierDomainExpression
from .normomegaexpr import NormAngularFourierDomainExpression
from .omegaexpr import AngularFourierDomainExpression
from .phasor import PhasorDomainExpression, PhasorRatioDomainExpression
from .sexpr import LaplaceDomainExpression
from .texpr import TimeDomainExpression
from .uexpr import UndefinedDomainExpression
from .zexpr import ZDomainExpression

__all__ = ('UndefinedDomainVoltage', 'UndefinedDomainCurrent', 'UndefinedDomainAdmittance', 'UndefinedDomainImpedance', 'UndefinedDomainTransferFunction', 'UndefinedDomainVoltagesquared', 'UndefinedDomainCurrentsquared', 'UndefinedDomainAdmittancesquared', 'UndefinedDomainImpedancesquared', 'UndefinedDomainPower', 'ConstantDomainVoltage', 'ConstantDomainCurrent', 'ConstantDomainAdmittance', 'ConstantDomainImpedance', 'ConstantDomainTransferFunction', 'ConstantDomainVoltagesquared', 'ConstantDomainCurrentsquared', 'ConstantDomainAdmittancesquared', 'ConstantDomainImpedancesquared', 'ConstantDomainPower', 'ConstantTimeDomainVoltage', 'ConstantTimeDomainCurrent', 'ConstantTimeDomainAdmittance', 'ConstantTimeDomainImpedance', 'ConstantTimeDomainTransferFunction', 'ConstantTimeDomainVoltagesquared', 'ConstantTimeDomainCurrentsquared', 'ConstantTimeDomainAdmittancesquared', 'ConstantTimeDomainImpedancesquared', 'ConstantTimeDomainPower', 'ConstantFrequencyResponseDomainVoltage', 'ConstantFrequencyResponseDomainCurrent', 'ConstantFrequencyResponseDomainAdmittance', 'ConstantFrequencyResponseDomainImpedance', 'ConstantFrequencyResponseDomainTransferFunction', 'ConstantFrequencyResponseDomainVoltagesquared', 'ConstantFrequencyResponseDomainCurrentsquared', 'ConstantFrequencyResponseDomainAdmittancesquared', 'ConstantFrequencyResponseDomainImpedancesquared', 'ConstantFrequencyResponseDomainPower', 'TimeDomainVoltage', 'TimeDomainCurrent', 'TimeDomainAdmittance', 'TimeDomainImpedance', 'TimeDomainTransferFunction', 'TimeDomainVoltagesquared', 'TimeDomainCurrentsquared', 'TimeDomainAdmittancesquared', 'TimeDomainImpedancesquared', 'TimeDomainPower', 'LaplaceDomainVoltage', 'LaplaceDomainCurrent', 'LaplaceDomainAdmittance', 'LaplaceDomainImpedance', 'LaplaceDomainTransferFunction', 'LaplaceDomainVoltagesquared', 'LaplaceDomainCurrentsquared', 'LaplaceDomainAdmittancesquared', 'LaplaceDomainImpedancesquared', 'LaplaceDomainPower', 'FourierDomainVoltage', 'FourierDomainCurrent', 'FourierDomainAdmittance', 'FourierDomainImpedance', 'FourierDomainTransferFunction', 'FourierDomainVoltagesquared', 'FourierDomainCurrentsquared', 'FourierDomainAdmittancesquared', 'FourierDomainImpedancesquared', 'FourierDomainPower', 'NormFourierDomainVoltage', 'NormFourierDomainCurrent', 'NormFourierDomainAdmittance', 'NormFourierDomainImpedance', 'NormFourierDomainTransferFunction', 'NormFourierDomainVoltagesquared', 'NormFourierDomainCurrentsquared', 'NormFourierDomainAdmittancesquared', 'NormFourierDomainImpedancesquared', 'NormFourierDomainPower', 'AngularFourierDomainVoltage', 'AngularFourierDomainCurrent', 'AngularFourierDomainAdmittance', 'AngularFourierDomainImpedance', 'AngularFourierDomainTransferFunction', 'AngularFourierDomainVoltagesquared', 'AngularFourierDomainCurrentsquared', 'AngularFourierDomainAdmittancesquared', 'AngularFourierDomainImpedancesquared', 'AngularFourierDomainPower', 'NormAngularFourierDomainVoltage', 'NormAngularFourierDomainCurrent', 'NormAngularFourierDomainAdmittance', 'NormAngularFourierDomainImpedance', 'NormAngularFourierDomainTransferFunction', 'NormAngularFourierDomainVoltagesquared', 'NormAngularFourierDomainCurrentsquared', 'NormAngularFourierDomainAdmittancesquared', 'NormAngularFourierDomainImpedancesquared', 'NormAngularFourierDomainPower', 'FrequencyResponseDomainVoltage', 'FrequencyResponseDomainCurrent', 'FrequencyResponseDomainAdmittance', 'FrequencyResponseDomainImpedance', 'FrequencyResponseDomainTransferFunction', 'FrequencyResponseDomainVoltagesquared', 'FrequencyResponseDomainCurrentsquared', 'FrequencyResponseDomainAdmittancesquared', 'FrequencyResponseDomainImpedancesquared', 'FrequencyResponseDomainPower', 'AngularFrequencyResponseDomainVoltage', 'AngularFrequencyResponseDomainCurrent', 'AngularFrequencyResponseDomainAdmittance', 'AngularFrequencyResponseDomainImpedance', 'AngularFrequencyResponseDomainTransferFunction', 'AngularFrequencyResponseDomainVoltagesquared', 'AngularFrequencyResponseDomainCurrentsquared', 'AngularFrequencyResponseDomainAdmittancesquared', 'AngularFrequencyResponseDomainImpedancesquared', 'AngularFrequencyResponseDomainPower', 'PhasorDomainVoltage', 'PhasorDomainCurrent', 'PhasorDomainAdmittance', 'PhasorDomainImpedance', 'PhasorDomainTransferFunction', 'PhasorDomainVoltagesquared', 'PhasorDomainCurrentsquared', 'PhasorDomainAdmittancesquared', 'PhasorDomainImpedancesquared', 'PhasorDomainPower', 'PhasorRatioDomainVoltage', 'PhasorRatioDomainCurrent', 'PhasorRatioDomainAdmittance', 'PhasorRatioDomainImpedance', 'PhasorRatioDomainTransferFunction', 'PhasorRatioDomainVoltagesquared', 'PhasorRatioDomainCurrentsquared', 'PhasorRatioDomainAdmittancesquared', 'PhasorRatioDomainImpedancesquared', 'PhasorRatioDomainPower', 'FourierNoiseDomainVoltage', 'FourierNoiseDomainCurrent', 'FourierNoiseDomainAdmittance', 'FourierNoiseDomainImpedance', 'FourierNoiseDomainTransferFunction', 'FourierNoiseDomainVoltagesquared', 'FourierNoiseDomainCurrentsquared', 'FourierNoiseDomainAdmittancesquared', 'FourierNoiseDomainImpedancesquared', 'FourierNoiseDomainPower', 'AngularFourierNoiseDomainVoltage', 'AngularFourierNoiseDomainCurrent', 'AngularFourierNoiseDomainAdmittance', 'AngularFourierNoiseDomainImpedance', 'AngularFourierNoiseDomainTransferFunction', 'AngularFourierNoiseDomainVoltagesquared', 'AngularFourierNoiseDomainCurrentsquared', 'AngularFourierNoiseDomainAdmittancesquared', 'AngularFourierNoiseDomainImpedancesquared', 'AngularFourierNoiseDomainPower', 'DiscreteTimeDomainVoltage', 'DiscreteTimeDomainCurrent', 'DiscreteTimeDomainAdmittance', 'DiscreteTimeDomainImpedance', 'DiscreteTimeDomainTransferFunction', 'DiscreteTimeDomainVoltagesquared', 'DiscreteTimeDomainCurrentsquared', 'DiscreteTimeDomainAdmittancesquared', 'DiscreteTimeDomainImpedancesquared', 'DiscreteTimeDomainPower', 'DiscreteFourierDomainVoltage', 'DiscreteFourierDomainCurrent', 'DiscreteFourierDomainAdmittance', 'DiscreteFourierDomainImpedance', 'DiscreteFourierDomainTransferFunction', 'DiscreteFourierDomainVoltagesquared', 'DiscreteFourierDomainCurrentsquared', 'DiscreteFourierDomainAdmittancesquared', 'DiscreteFourierDomainImpedancesquared', 'DiscreteFourierDomainPower', 'ZDomainVoltage', 'ZDomainCurrent', 'ZDomainAdmittance', 'ZDomainImpedance', 'ZDomainTransferFunction', 'ZDomainVoltagesquared', 'ZDomainCurrentsquared', 'ZDomainAdmittancesquared', 'ZDomainImpedancesquared', 'ZDomainPower')



class UndefinedDomainVoltage(VoltageMixin, UndefinedDomainExpression):
    """Undefined-domain voltage (units V)."""

    _default_units = uu.volt


class UndefinedDomainCurrent(CurrentMixin, UndefinedDomainExpression):
    """Undefined-domain current (units A)."""

    _default_units = uu.ampere


class UndefinedDomainAdmittance(AdmittanceMixin, UndefinedDomainExpression):
    """Undefined-domain admittance (units S)."""

    _default_units = uu.siemens


class UndefinedDomainImpedance(ImpedanceMixin, UndefinedDomainExpression):
    """Undefined-domain impedance (units ohm)."""

    _default_units = uu.ohm


class UndefinedDomainTransferFunction(TransferMixin, UndefinedDomainExpression):
    """Undefined-domain transfer (units )."""

    _default_units = S.One


class UndefinedDomainVoltagesquared(VoltageSquaredMixin, UndefinedDomainExpression):
    """Undefined-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class UndefinedDomainCurrentsquared(CurrentSquaredMixin, UndefinedDomainExpression):
    """Undefined-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class UndefinedDomainAdmittancesquared(AdmittanceSquaredMixin, UndefinedDomainExpression):
    """Undefined-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class UndefinedDomainImpedancesquared(ImpedanceSquaredMixin, UndefinedDomainExpression):
    """Undefined-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class UndefinedDomainPower(PowerMixin, UndefinedDomainExpression):
    """Undefined-domain power (units W)."""

    _default_units = uu.watt


class ConstantDomainVoltage(VoltageMixin, ConstantDomainExpression):
    """-domain voltage (units V)."""

    _default_units = uu.volt


class ConstantDomainCurrent(CurrentMixin, ConstantDomainExpression):
    """-domain current (units A)."""

    _default_units = uu.ampere


class ConstantDomainAdmittance(AdmittanceMixin, ConstantDomainExpression):
    """-domain admittance (units S)."""

    _default_units = uu.siemens


class ConstantDomainImpedance(ImpedanceMixin, ConstantDomainExpression):
    """-domain impedance (units ohm)."""

    _default_units = uu.ohm


class ConstantDomainTransferFunction(TransferMixin, ConstantDomainExpression):
    """-domain transfer (units )."""

    _default_units = S.One


class ConstantDomainVoltagesquared(VoltageSquaredMixin, ConstantDomainExpression):
    """-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class ConstantDomainCurrentsquared(CurrentSquaredMixin, ConstantDomainExpression):
    """-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class ConstantDomainAdmittancesquared(AdmittanceSquaredMixin, ConstantDomainExpression):
    """-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class ConstantDomainImpedancesquared(ImpedanceSquaredMixin, ConstantDomainExpression):
    """-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class ConstantDomainPower(PowerMixin, ConstantDomainExpression):
    """-domain power (units W)."""

    _default_units = uu.watt


class ConstantTimeDomainVoltage(VoltageMixin, ConstantTimeDomainExpression):
    """-domain voltage (units V)."""

    _default_units = uu.volt


class ConstantTimeDomainCurrent(CurrentMixin, ConstantTimeDomainExpression):
    """-domain current (units A)."""

    _default_units = uu.ampere


class ConstantTimeDomainAdmittance(AdmittanceMixin, ConstantTimeDomainExpression):
    """-domain admittance (units S)."""

    _default_units = uu.siemens


class ConstantTimeDomainImpedance(ImpedanceMixin, ConstantTimeDomainExpression):
    """-domain impedance (units ohm)."""

    _default_units = uu.ohm


class ConstantTimeDomainTransferFunction(TransferMixin, ConstantTimeDomainExpression):
    """-domain transfer (units )."""

    _default_units = S.One


class ConstantTimeDomainVoltagesquared(VoltageSquaredMixin, ConstantTimeDomainExpression):
    """-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class ConstantTimeDomainCurrentsquared(CurrentSquaredMixin, ConstantTimeDomainExpression):
    """-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class ConstantTimeDomainAdmittancesquared(AdmittanceSquaredMixin, ConstantTimeDomainExpression):
    """-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class ConstantTimeDomainImpedancesquared(ImpedanceSquaredMixin, ConstantTimeDomainExpression):
    """-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class ConstantTimeDomainPower(PowerMixin, ConstantTimeDomainExpression):
    """-domain power (units W)."""

    _default_units = uu.watt


class ConstantFrequencyResponseDomainVoltage(VoltageMixin, ConstantFrequencyResponseDomainExpression):
    """-domain voltage (units V)."""

    _default_units = uu.volt


class ConstantFrequencyResponseDomainCurrent(CurrentMixin, ConstantFrequencyResponseDomainExpression):
    """-domain current (units A)."""

    _default_units = uu.ampere


class ConstantFrequencyResponseDomainAdmittance(AdmittanceMixin, ConstantFrequencyResponseDomainExpression):
    """-domain admittance (units S)."""

    _default_units = uu.siemens


class ConstantFrequencyResponseDomainImpedance(ImpedanceMixin, ConstantFrequencyResponseDomainExpression):
    """-domain impedance (units ohm)."""

    _default_units = uu.ohm


class ConstantFrequencyResponseDomainTransferFunction(TransferMixin, ConstantFrequencyResponseDomainExpression):
    """-domain transfer (units )."""

    _default_units = S.One


class ConstantFrequencyResponseDomainVoltagesquared(VoltageSquaredMixin, ConstantFrequencyResponseDomainExpression):
    """-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class ConstantFrequencyResponseDomainCurrentsquared(CurrentSquaredMixin, ConstantFrequencyResponseDomainExpression):
    """-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class ConstantFrequencyResponseDomainAdmittancesquared(AdmittanceSquaredMixin, ConstantFrequencyResponseDomainExpression):
    """-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class ConstantFrequencyResponseDomainImpedancesquared(ImpedanceSquaredMixin, ConstantFrequencyResponseDomainExpression):
    """-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class ConstantFrequencyResponseDomainPower(PowerMixin, ConstantFrequencyResponseDomainExpression):
    """-domain power (units W)."""

    _default_units = uu.watt


class TimeDomainVoltage(VoltageMixin, TimeDomainExpression):
    """Time-domain voltage (units V)."""

    _default_units = uu.volt


class TimeDomainCurrent(CurrentMixin, TimeDomainExpression):
    """Time-domain current (units A)."""

    _default_units = uu.ampere


class TimeDomainAdmittance(AdmittanceMixin, TimeDomainExpression):
    """Time-domain admittance (units S/s)."""

    _default_units = uu.siemens / uu.s


class TimeDomainImpedance(ImpedanceMixin, TimeDomainExpression):
    """Time-domain impedance (units ohm/s)."""

    _default_units = uu.ohm / uu.s


class TimeDomainTransferFunction(TransferMixin, TimeDomainExpression):
    """Time-domain transfer (units /s)."""

    _default_units = 1 / uu.s


class TimeDomainVoltagesquared(VoltageSquaredMixin, TimeDomainExpression):
    """Time-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class TimeDomainCurrentsquared(CurrentSquaredMixin, TimeDomainExpression):
    """Time-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class TimeDomainAdmittancesquared(AdmittanceSquaredMixin, TimeDomainExpression):
    """Time-domain admittancesquared (units S^2/s^2)."""

    _default_units = (uu.siemens / uu.s)**2


class TimeDomainImpedancesquared(ImpedanceSquaredMixin, TimeDomainExpression):
    """Time-domain impedancesquared (units ohm^2/s^2)."""

    _default_units = (uu.ohm / uu.s)**2


class TimeDomainPower(PowerMixin, TimeDomainExpression):
    """Time-domain power (units W)."""

    _default_units = uu.watt


class LaplaceDomainVoltage(VoltageMixin, LaplaceDomainExpression):
    """Angular frequency-domain voltage (units V/Hz)."""

    _default_units = uu.volt / uu.Hz


class LaplaceDomainCurrent(CurrentMixin, LaplaceDomainExpression):
    """Angular frequency-domain current (units A/Hz)."""

    _default_units = uu.ampere / uu.Hz


class LaplaceDomainAdmittance(AdmittanceMixin, LaplaceDomainExpression):
    """Angular frequency-domain admittance (units S)."""

    _default_units = uu.siemens


class LaplaceDomainImpedance(ImpedanceMixin, LaplaceDomainExpression):
    """Angular frequency-domain impedance (units ohm)."""

    _default_units = uu.ohm


class LaplaceDomainTransferFunction(TransferMixin, LaplaceDomainExpression):
    """Angular frequency-domain transfer (units )."""

    _default_units = S.One


class LaplaceDomainVoltagesquared(VoltageSquaredMixin, LaplaceDomainExpression):
    """Angular frequency-domain voltagesquared (units V^2/Hz^2)."""

    _default_units = (uu.volt / uu.Hz)**2


class LaplaceDomainCurrentsquared(CurrentSquaredMixin, LaplaceDomainExpression):
    """Angular frequency-domain currentsquared (units A^2/Hz^2)."""

    _default_units = (uu.ampere / uu.Hz)**2


class LaplaceDomainAdmittancesquared(AdmittanceSquaredMixin, LaplaceDomainExpression):
    """Angular frequency-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class LaplaceDomainImpedancesquared(ImpedanceSquaredMixin, LaplaceDomainExpression):
    """Angular frequency-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class LaplaceDomainPower(PowerMixin, LaplaceDomainExpression):
    """Angular frequency-domain power (units W)."""

    _default_units = uu.watt


class FourierDomainVoltage(VoltageMixin, FourierDomainExpression):
    """Frequency-domain voltage (units V/Hz)."""

    _default_units = uu.volt / uu.Hz


class FourierDomainCurrent(CurrentMixin, FourierDomainExpression):
    """Frequency-domain current (units A/Hz)."""

    _default_units = uu.ampere / uu.Hz


class FourierDomainAdmittance(AdmittanceMixin, FourierDomainExpression):
    """Frequency-domain admittance (units S)."""

    _default_units = uu.siemens


class FourierDomainImpedance(ImpedanceMixin, FourierDomainExpression):
    """Frequency-domain impedance (units ohm)."""

    _default_units = uu.ohm


class FourierDomainTransferFunction(TransferMixin, FourierDomainExpression):
    """Frequency-domain transfer (units )."""

    _default_units = S.One


class FourierDomainVoltagesquared(VoltageSquaredMixin, FourierDomainExpression):
    """Frequency-domain voltagesquared (units V^2/Hz^2)."""

    _default_units = (uu.volt / uu.Hz)**2


class FourierDomainCurrentsquared(CurrentSquaredMixin, FourierDomainExpression):
    """Frequency-domain currentsquared (units A^2/Hz^2)."""

    _default_units = (uu.ampere / uu.Hz)**2


class FourierDomainAdmittancesquared(AdmittanceSquaredMixin, FourierDomainExpression):
    """Frequency-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class FourierDomainImpedancesquared(ImpedanceSquaredMixin, FourierDomainExpression):
    """Frequency-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class FourierDomainPower(PowerMixin, FourierDomainExpression):
    """Frequency-domain power (units W)."""

    _default_units = uu.watt


class NormFourierDomainVoltage(VoltageMixin, NormFourierDomainExpression):
    """Normalized frequency-domain voltage (units V)."""

    _default_units = uu.volt


class NormFourierDomainCurrent(CurrentMixin, NormFourierDomainExpression):
    """Normalized frequency-domain current (units A)."""

    _default_units = uu.ampere


class NormFourierDomainAdmittance(AdmittanceMixin, NormFourierDomainExpression):
    """Normalized frequency-domain admittance (units S)."""

    _default_units = uu.siemens


class NormFourierDomainImpedance(ImpedanceMixin, NormFourierDomainExpression):
    """Normalized frequency-domain impedance (units ohm)."""

    _default_units = uu.ohm


class NormFourierDomainTransferFunction(TransferMixin, NormFourierDomainExpression):
    """Normalized frequency-domain transfer (units )."""

    _default_units = S.One


class NormFourierDomainVoltagesquared(VoltageSquaredMixin, NormFourierDomainExpression):
    """Normalized frequency-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class NormFourierDomainCurrentsquared(CurrentSquaredMixin, NormFourierDomainExpression):
    """Normalized frequency-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class NormFourierDomainAdmittancesquared(AdmittanceSquaredMixin, NormFourierDomainExpression):
    """Normalized frequency-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class NormFourierDomainImpedancesquared(ImpedanceSquaredMixin, NormFourierDomainExpression):
    """Normalized frequency-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class NormFourierDomainPower(PowerMixin, NormFourierDomainExpression):
    """Normalized frequency-domain power (units W)."""

    _default_units = uu.watt


class AngularFourierDomainVoltage(VoltageMixin, AngularFourierDomainExpression):
    """Angular frequency-domain voltage (units V/Hz)."""

    _default_units = uu.volt / uu.Hz


class AngularFourierDomainCurrent(CurrentMixin, AngularFourierDomainExpression):
    """Angular frequency-domain current (units A/Hz)."""

    _default_units = uu.ampere / uu.Hz


class AngularFourierDomainAdmittance(AdmittanceMixin, AngularFourierDomainExpression):
    """Angular frequency-domain admittance (units S)."""

    _default_units = uu.siemens


class AngularFourierDomainImpedance(ImpedanceMixin, AngularFourierDomainExpression):
    """Angular frequency-domain impedance (units ohm)."""

    _default_units = uu.ohm


class AngularFourierDomainTransferFunction(TransferMixin, AngularFourierDomainExpression):
    """Angular frequency-domain transfer (units )."""

    _default_units = S.One


class AngularFourierDomainVoltagesquared(VoltageSquaredMixin, AngularFourierDomainExpression):
    """Angular frequency-domain voltagesquared (units V^2/Hz^2)."""

    _default_units = (uu.volt / uu.Hz)**2


class AngularFourierDomainCurrentsquared(CurrentSquaredMixin, AngularFourierDomainExpression):
    """Angular frequency-domain currentsquared (units A^2/Hz^2)."""

    _default_units = (uu.ampere / uu.Hz)**2


class AngularFourierDomainAdmittancesquared(AdmittanceSquaredMixin, AngularFourierDomainExpression):
    """Angular frequency-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class AngularFourierDomainImpedancesquared(ImpedanceSquaredMixin, AngularFourierDomainExpression):
    """Angular frequency-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class AngularFourierDomainPower(PowerMixin, AngularFourierDomainExpression):
    """Angular frequency-domain power (units W)."""

    _default_units = uu.watt


class NormAngularFourierDomainVoltage(VoltageMixin, NormAngularFourierDomainExpression):
    """Normalized angular frequency-domain voltage (units V)."""

    _default_units = uu.volt


class NormAngularFourierDomainCurrent(CurrentMixin, NormAngularFourierDomainExpression):
    """Normalized angular frequency-domain current (units A)."""

    _default_units = uu.ampere


class NormAngularFourierDomainAdmittance(AdmittanceMixin, NormAngularFourierDomainExpression):
    """Normalized angular frequency-domain admittance (units S)."""

    _default_units = uu.siemens


class NormAngularFourierDomainImpedance(ImpedanceMixin, NormAngularFourierDomainExpression):
    """Normalized angular frequency-domain impedance (units ohm)."""

    _default_units = uu.ohm


class NormAngularFourierDomainTransferFunction(TransferMixin, NormAngularFourierDomainExpression):
    """Normalized angular frequency-domain transfer (units )."""

    _default_units = S.One


class NormAngularFourierDomainVoltagesquared(VoltageSquaredMixin, NormAngularFourierDomainExpression):
    """Normalized angular frequency-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class NormAngularFourierDomainCurrentsquared(CurrentSquaredMixin, NormAngularFourierDomainExpression):
    """Normalized angular frequency-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class NormAngularFourierDomainAdmittancesquared(AdmittanceSquaredMixin, NormAngularFourierDomainExpression):
    """Normalized angular frequency-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class NormAngularFourierDomainImpedancesquared(ImpedanceSquaredMixin, NormAngularFourierDomainExpression):
    """Normalized angular frequency-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class NormAngularFourierDomainPower(PowerMixin, NormAngularFourierDomainExpression):
    """Normalized angular frequency-domain power (units W)."""

    _default_units = uu.watt


class FrequencyResponseDomainVoltage(VoltageMixin, FrequencyResponseDomainExpression):
    """Frequency-domain voltage (units V)."""

    _default_units = uu.volt


class FrequencyResponseDomainCurrent(CurrentMixin, FrequencyResponseDomainExpression):
    """Frequency-domain current (units A)."""

    _default_units = uu.ampere


class FrequencyResponseDomainAdmittance(AdmittanceMixin, FrequencyResponseDomainExpression):
    """Frequency-domain admittance (units S)."""

    _default_units = uu.siemens


class FrequencyResponseDomainImpedance(ImpedanceMixin, FrequencyResponseDomainExpression):
    """Frequency-domain impedance (units ohm)."""

    _default_units = uu.ohm


class FrequencyResponseDomainTransferFunction(TransferMixin, FrequencyResponseDomainExpression):
    """Frequency-domain transfer (units )."""

    _default_units = S.One


class FrequencyResponseDomainVoltagesquared(VoltageSquaredMixin, FrequencyResponseDomainExpression):
    """Frequency-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class FrequencyResponseDomainCurrentsquared(CurrentSquaredMixin, FrequencyResponseDomainExpression):
    """Frequency-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class FrequencyResponseDomainAdmittancesquared(AdmittanceSquaredMixin, FrequencyResponseDomainExpression):
    """Frequency-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class FrequencyResponseDomainImpedancesquared(ImpedanceSquaredMixin, FrequencyResponseDomainExpression):
    """Frequency-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class FrequencyResponseDomainPower(PowerMixin, FrequencyResponseDomainExpression):
    """Frequency-domain power (units W)."""

    _default_units = uu.watt


class AngularFrequencyResponseDomainVoltage(VoltageMixin, AngularFrequencyResponseDomainExpression):
    """Angular frequency-domain voltage (units V)."""

    _default_units = uu.volt


class AngularFrequencyResponseDomainCurrent(CurrentMixin, AngularFrequencyResponseDomainExpression):
    """Angular frequency-domain current (units A)."""

    _default_units = uu.ampere


class AngularFrequencyResponseDomainAdmittance(AdmittanceMixin, AngularFrequencyResponseDomainExpression):
    """Angular frequency-domain admittance (units S)."""

    _default_units = uu.siemens


class AngularFrequencyResponseDomainImpedance(ImpedanceMixin, AngularFrequencyResponseDomainExpression):
    """Angular frequency-domain impedance (units ohm)."""

    _default_units = uu.ohm


class AngularFrequencyResponseDomainTransferFunction(TransferMixin, AngularFrequencyResponseDomainExpression):
    """Angular frequency-domain transfer (units )."""

    _default_units = S.One


class AngularFrequencyResponseDomainVoltagesquared(VoltageSquaredMixin, AngularFrequencyResponseDomainExpression):
    """Angular frequency-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class AngularFrequencyResponseDomainCurrentsquared(CurrentSquaredMixin, AngularFrequencyResponseDomainExpression):
    """Angular frequency-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class AngularFrequencyResponseDomainAdmittancesquared(AdmittanceSquaredMixin, AngularFrequencyResponseDomainExpression):
    """Angular frequency-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class AngularFrequencyResponseDomainImpedancesquared(ImpedanceSquaredMixin, AngularFrequencyResponseDomainExpression):
    """Angular frequency-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class AngularFrequencyResponseDomainPower(PowerMixin, AngularFrequencyResponseDomainExpression):
    """Angular frequency-domain power (units W)."""

    _default_units = uu.watt


class PhasorDomainVoltage(VoltageMixin, PhasorDomainExpression):
    """-domain voltage (units V)."""

    _default_units = uu.volt


class PhasorDomainCurrent(CurrentMixin, PhasorDomainExpression):
    """-domain current (units A)."""

    _default_units = uu.ampere


class PhasorDomainAdmittance(AdmittanceMixin, PhasorDomainExpression):
    """-domain admittance (units S)."""

    _default_units = uu.siemens


class PhasorDomainImpedance(ImpedanceMixin, PhasorDomainExpression):
    """-domain impedance (units ohm)."""

    _default_units = uu.ohm


class PhasorDomainTransferFunction(TransferMixin, PhasorDomainExpression):
    """-domain transfer (units )."""

    _default_units = S.One


class PhasorDomainVoltagesquared(VoltageSquaredMixin, PhasorDomainExpression):
    """-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class PhasorDomainCurrentsquared(CurrentSquaredMixin, PhasorDomainExpression):
    """-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class PhasorDomainAdmittancesquared(AdmittanceSquaredMixin, PhasorDomainExpression):
    """-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class PhasorDomainImpedancesquared(ImpedanceSquaredMixin, PhasorDomainExpression):
    """-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class PhasorDomainPower(PowerMixin, PhasorDomainExpression):
    """-domain power (units W)."""

    _default_units = uu.watt


class PhasorRatioDomainVoltage(VoltageMixin, PhasorRatioDomainExpression):
    """Angular frequency-domain voltage (units V)."""

    _default_units = uu.volt


class PhasorRatioDomainCurrent(CurrentMixin, PhasorRatioDomainExpression):
    """Angular frequency-domain current (units A)."""

    _default_units = uu.ampere


class PhasorRatioDomainAdmittance(AdmittanceMixin, PhasorRatioDomainExpression):
    """Angular frequency-domain admittance (units S)."""

    _default_units = uu.siemens


class PhasorRatioDomainImpedance(ImpedanceMixin, PhasorRatioDomainExpression):
    """Angular frequency-domain impedance (units ohm)."""

    _default_units = uu.ohm


class PhasorRatioDomainTransferFunction(TransferMixin, PhasorRatioDomainExpression):
    """Angular frequency-domain transfer (units )."""

    _default_units = S.One


class PhasorRatioDomainVoltagesquared(VoltageSquaredMixin, PhasorRatioDomainExpression):
    """Angular frequency-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class PhasorRatioDomainCurrentsquared(CurrentSquaredMixin, PhasorRatioDomainExpression):
    """Angular frequency-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class PhasorRatioDomainAdmittancesquared(AdmittanceSquaredMixin, PhasorRatioDomainExpression):
    """Angular frequency-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class PhasorRatioDomainImpedancesquared(ImpedanceSquaredMixin, PhasorRatioDomainExpression):
    """Angular frequency-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class PhasorRatioDomainPower(PowerMixin, PhasorRatioDomainExpression):
    """Angular frequency-domain power (units W)."""

    _default_units = uu.watt


class FourierNoiseDomainVoltage(VoltageMixin, FourierNoiseDomainExpression):
    """Frequency-domain voltage (units V)."""

    _default_units = uu.volt


class FourierNoiseDomainCurrent(CurrentMixin, FourierNoiseDomainExpression):
    """Frequency-domain current (units A)."""

    _default_units = uu.ampere


class FourierNoiseDomainAdmittance(AdmittanceMixin, FourierNoiseDomainExpression):
    """Frequency-domain admittance (units S)."""

    _default_units = uu.siemens


class FourierNoiseDomainImpedance(ImpedanceMixin, FourierNoiseDomainExpression):
    """Frequency-domain impedance (units ohm)."""

    _default_units = uu.ohm


class FourierNoiseDomainTransferFunction(TransferMixin, FourierNoiseDomainExpression):
    """Frequency-domain transfer (units )."""

    _default_units = S.One


class FourierNoiseDomainVoltagesquared(VoltageSquaredMixin, FourierNoiseDomainExpression):
    """Frequency-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class FourierNoiseDomainCurrentsquared(CurrentSquaredMixin, FourierNoiseDomainExpression):
    """Frequency-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class FourierNoiseDomainAdmittancesquared(AdmittanceSquaredMixin, FourierNoiseDomainExpression):
    """Frequency-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class FourierNoiseDomainImpedancesquared(ImpedanceSquaredMixin, FourierNoiseDomainExpression):
    """Frequency-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class FourierNoiseDomainPower(PowerMixin, FourierNoiseDomainExpression):
    """Frequency-domain power (units W)."""

    _default_units = uu.watt


class AngularFourierNoiseDomainVoltage(VoltageMixin, AngularFourierNoiseDomainExpression):
    """Angular frequency-domain voltage (units V)."""

    _default_units = uu.volt


class AngularFourierNoiseDomainCurrent(CurrentMixin, AngularFourierNoiseDomainExpression):
    """Angular frequency-domain current (units A)."""

    _default_units = uu.ampere


class AngularFourierNoiseDomainAdmittance(AdmittanceMixin, AngularFourierNoiseDomainExpression):
    """Angular frequency-domain admittance (units S)."""

    _default_units = uu.siemens


class AngularFourierNoiseDomainImpedance(ImpedanceMixin, AngularFourierNoiseDomainExpression):
    """Angular frequency-domain impedance (units ohm)."""

    _default_units = uu.ohm


class AngularFourierNoiseDomainTransferFunction(TransferMixin, AngularFourierNoiseDomainExpression):
    """Angular frequency-domain transfer (units )."""

    _default_units = S.One


class AngularFourierNoiseDomainVoltagesquared(VoltageSquaredMixin, AngularFourierNoiseDomainExpression):
    """Angular frequency-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class AngularFourierNoiseDomainCurrentsquared(CurrentSquaredMixin, AngularFourierNoiseDomainExpression):
    """Angular frequency-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class AngularFourierNoiseDomainAdmittancesquared(AdmittanceSquaredMixin, AngularFourierNoiseDomainExpression):
    """Angular frequency-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class AngularFourierNoiseDomainImpedancesquared(ImpedanceSquaredMixin, AngularFourierNoiseDomainExpression):
    """Angular frequency-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class AngularFourierNoiseDomainPower(PowerMixin, AngularFourierNoiseDomainExpression):
    """Angular frequency-domain power (units W)."""

    _default_units = uu.watt


class DiscreteTimeDomainVoltage(VoltageMixin, DiscreteTimeDomainExpression):
    """Discrete time-domain voltage (units V)."""

    _default_units = uu.volt


class DiscreteTimeDomainCurrent(CurrentMixin, DiscreteTimeDomainExpression):
    """Discrete time-domain current (units A)."""

    _default_units = uu.ampere


class DiscreteTimeDomainAdmittance(AdmittanceMixin, DiscreteTimeDomainExpression):
    """Discrete time-domain admittance (units S)."""

    _default_units = uu.siemens


class DiscreteTimeDomainImpedance(ImpedanceMixin, DiscreteTimeDomainExpression):
    """Discrete time-domain impedance (units ohm)."""

    _default_units = uu.ohm


class DiscreteTimeDomainTransferFunction(TransferMixin, DiscreteTimeDomainExpression):
    """Discrete time-domain transfer (units )."""

    _default_units = S.One


class DiscreteTimeDomainVoltagesquared(VoltageSquaredMixin, DiscreteTimeDomainExpression):
    """Discrete time-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class DiscreteTimeDomainCurrentsquared(CurrentSquaredMixin, DiscreteTimeDomainExpression):
    """Discrete time-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class DiscreteTimeDomainAdmittancesquared(AdmittanceSquaredMixin, DiscreteTimeDomainExpression):
    """Discrete time-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class DiscreteTimeDomainImpedancesquared(ImpedanceSquaredMixin, DiscreteTimeDomainExpression):
    """Discrete time-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class DiscreteTimeDomainPower(PowerMixin, DiscreteTimeDomainExpression):
    """Discrete time-domain power (units W)."""

    _default_units = uu.watt


class DiscreteFourierDomainVoltage(VoltageMixin, DiscreteFourierDomainExpression):
    """Discrete frequency-domain voltage (units V)."""

    _default_units = uu.volt


class DiscreteFourierDomainCurrent(CurrentMixin, DiscreteFourierDomainExpression):
    """Discrete frequency-domain current (units A)."""

    _default_units = uu.ampere


class DiscreteFourierDomainAdmittance(AdmittanceMixin, DiscreteFourierDomainExpression):
    """Discrete frequency-domain admittance (units S)."""

    _default_units = uu.siemens


class DiscreteFourierDomainImpedance(ImpedanceMixin, DiscreteFourierDomainExpression):
    """Discrete frequency-domain impedance (units ohm)."""

    _default_units = uu.ohm


class DiscreteFourierDomainTransferFunction(TransferMixin, DiscreteFourierDomainExpression):
    """Discrete frequency-domain transfer (units )."""

    _default_units = S.One


class DiscreteFourierDomainVoltagesquared(VoltageSquaredMixin, DiscreteFourierDomainExpression):
    """Discrete frequency-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class DiscreteFourierDomainCurrentsquared(CurrentSquaredMixin, DiscreteFourierDomainExpression):
    """Discrete frequency-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class DiscreteFourierDomainAdmittancesquared(AdmittanceSquaredMixin, DiscreteFourierDomainExpression):
    """Discrete frequency-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class DiscreteFourierDomainImpedancesquared(ImpedanceSquaredMixin, DiscreteFourierDomainExpression):
    """Discrete frequency-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class DiscreteFourierDomainPower(PowerMixin, DiscreteFourierDomainExpression):
    """Discrete frequency-domain power (units W)."""

    _default_units = uu.watt


class ZDomainVoltage(VoltageMixin, ZDomainExpression):
    """Z-domain voltage (units V)."""

    _default_units = uu.volt


class ZDomainCurrent(CurrentMixin, ZDomainExpression):
    """Z-domain current (units A)."""

    _default_units = uu.ampere


class ZDomainAdmittance(AdmittanceMixin, ZDomainExpression):
    """Z-domain admittance (units S)."""

    _default_units = uu.siemens


class ZDomainImpedance(ImpedanceMixin, ZDomainExpression):
    """Z-domain impedance (units ohm)."""

    _default_units = uu.ohm


class ZDomainTransferFunction(TransferMixin, ZDomainExpression):
    """Z-domain transfer (units )."""

    _default_units = S.One


class ZDomainVoltagesquared(VoltageSquaredMixin, ZDomainExpression):
    """Z-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class ZDomainCurrentsquared(CurrentSquaredMixin, ZDomainExpression):
    """Z-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class ZDomainAdmittancesquared(AdmittanceSquaredMixin, ZDomainExpression):
    """Z-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class ZDomainImpedancesquared(ImpedanceSquaredMixin, ZDomainExpression):
    """Z-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class ZDomainPower(PowerMixin, ZDomainExpression):
    """Z-domain power (units W)."""

    _default_units = uu.watt

exprclasses = {'undefined': {'undefined': UndefinedDomainExpression,
'voltage': UndefinedDomainVoltage,
'current': UndefinedDomainCurrent,
'admittance': UndefinedDomainAdmittance,
'impedance': UndefinedDomainImpedance,
'transfer': UndefinedDomainTransferFunction,
'voltagesquared': UndefinedDomainVoltagesquared,
'currentsquared': UndefinedDomainCurrentsquared,
'admittancesquared': UndefinedDomainAdmittancesquared,
'impedancesquared': UndefinedDomainImpedancesquared,
'power': UndefinedDomainPower,
}
,'constant': {'undefined': ConstantDomainExpression,
'voltage': ConstantDomainVoltage,
'current': ConstantDomainCurrent,
'admittance': ConstantDomainAdmittance,
'impedance': ConstantDomainImpedance,
'transfer': ConstantDomainTransferFunction,
'voltagesquared': ConstantDomainVoltagesquared,
'currentsquared': ConstantDomainCurrentsquared,
'admittancesquared': ConstantDomainAdmittancesquared,
'impedancesquared': ConstantDomainImpedancesquared,
'power': ConstantDomainPower,
}
,'constant time': {'undefined': ConstantTimeDomainExpression,
'voltage': ConstantTimeDomainVoltage,
'current': ConstantTimeDomainCurrent,
'admittance': ConstantTimeDomainAdmittance,
'impedance': ConstantTimeDomainImpedance,
'transfer': ConstantTimeDomainTransferFunction,
'voltagesquared': ConstantTimeDomainVoltagesquared,
'currentsquared': ConstantTimeDomainCurrentsquared,
'admittancesquared': ConstantTimeDomainAdmittancesquared,
'impedancesquared': ConstantTimeDomainImpedancesquared,
'power': ConstantTimeDomainPower,
}
,'constant frequency response': {'undefined': ConstantFrequencyResponseDomainExpression,
'voltage': ConstantFrequencyResponseDomainVoltage,
'current': ConstantFrequencyResponseDomainCurrent,
'admittance': ConstantFrequencyResponseDomainAdmittance,
'impedance': ConstantFrequencyResponseDomainImpedance,
'transfer': ConstantFrequencyResponseDomainTransferFunction,
'voltagesquared': ConstantFrequencyResponseDomainVoltagesquared,
'currentsquared': ConstantFrequencyResponseDomainCurrentsquared,
'admittancesquared': ConstantFrequencyResponseDomainAdmittancesquared,
'impedancesquared': ConstantFrequencyResponseDomainImpedancesquared,
'power': ConstantFrequencyResponseDomainPower,
}
,'time': {'undefined': TimeDomainExpression,
'voltage': TimeDomainVoltage,
'current': TimeDomainCurrent,
'admittance': TimeDomainAdmittance,
'impedance': TimeDomainImpedance,
'transfer': TimeDomainTransferFunction,
'voltagesquared': TimeDomainVoltagesquared,
'currentsquared': TimeDomainCurrentsquared,
'admittancesquared': TimeDomainAdmittancesquared,
'impedancesquared': TimeDomainImpedancesquared,
'power': TimeDomainPower,
}
,'laplace': {'undefined': LaplaceDomainExpression,
'voltage': LaplaceDomainVoltage,
'current': LaplaceDomainCurrent,
'admittance': LaplaceDomainAdmittance,
'impedance': LaplaceDomainImpedance,
'transfer': LaplaceDomainTransferFunction,
'voltagesquared': LaplaceDomainVoltagesquared,
'currentsquared': LaplaceDomainCurrentsquared,
'admittancesquared': LaplaceDomainAdmittancesquared,
'impedancesquared': LaplaceDomainImpedancesquared,
'power': LaplaceDomainPower,
}
,'fourier': {'undefined': FourierDomainExpression,
'voltage': FourierDomainVoltage,
'current': FourierDomainCurrent,
'admittance': FourierDomainAdmittance,
'impedance': FourierDomainImpedance,
'transfer': FourierDomainTransferFunction,
'voltagesquared': FourierDomainVoltagesquared,
'currentsquared': FourierDomainCurrentsquared,
'admittancesquared': FourierDomainAdmittancesquared,
'impedancesquared': FourierDomainImpedancesquared,
'power': FourierDomainPower,
}
,'norm fourier': {'undefined': NormFourierDomainExpression,
'voltage': NormFourierDomainVoltage,
'current': NormFourierDomainCurrent,
'admittance': NormFourierDomainAdmittance,
'impedance': NormFourierDomainImpedance,
'transfer': NormFourierDomainTransferFunction,
'voltagesquared': NormFourierDomainVoltagesquared,
'currentsquared': NormFourierDomainCurrentsquared,
'admittancesquared': NormFourierDomainAdmittancesquared,
'impedancesquared': NormFourierDomainImpedancesquared,
'power': NormFourierDomainPower,
}
,'angular fourier': {'undefined': AngularFourierDomainExpression,
'voltage': AngularFourierDomainVoltage,
'current': AngularFourierDomainCurrent,
'admittance': AngularFourierDomainAdmittance,
'impedance': AngularFourierDomainImpedance,
'transfer': AngularFourierDomainTransferFunction,
'voltagesquared': AngularFourierDomainVoltagesquared,
'currentsquared': AngularFourierDomainCurrentsquared,
'admittancesquared': AngularFourierDomainAdmittancesquared,
'impedancesquared': AngularFourierDomainImpedancesquared,
'power': AngularFourierDomainPower,
}
,'norm angular fourier': {'undefined': NormAngularFourierDomainExpression,
'voltage': NormAngularFourierDomainVoltage,
'current': NormAngularFourierDomainCurrent,
'admittance': NormAngularFourierDomainAdmittance,
'impedance': NormAngularFourierDomainImpedance,
'transfer': NormAngularFourierDomainTransferFunction,
'voltagesquared': NormAngularFourierDomainVoltagesquared,
'currentsquared': NormAngularFourierDomainCurrentsquared,
'admittancesquared': NormAngularFourierDomainAdmittancesquared,
'impedancesquared': NormAngularFourierDomainImpedancesquared,
'power': NormAngularFourierDomainPower,
}
,'frequency response': {'undefined': FrequencyResponseDomainExpression,
'voltage': FrequencyResponseDomainVoltage,
'current': FrequencyResponseDomainCurrent,
'admittance': FrequencyResponseDomainAdmittance,
'impedance': FrequencyResponseDomainImpedance,
'transfer': FrequencyResponseDomainTransferFunction,
'voltagesquared': FrequencyResponseDomainVoltagesquared,
'currentsquared': FrequencyResponseDomainCurrentsquared,
'admittancesquared': FrequencyResponseDomainAdmittancesquared,
'impedancesquared': FrequencyResponseDomainImpedancesquared,
'power': FrequencyResponseDomainPower,
}
,'angular frequency response': {'undefined': AngularFrequencyResponseDomainExpression,
'voltage': AngularFrequencyResponseDomainVoltage,
'current': AngularFrequencyResponseDomainCurrent,
'admittance': AngularFrequencyResponseDomainAdmittance,
'impedance': AngularFrequencyResponseDomainImpedance,
'transfer': AngularFrequencyResponseDomainTransferFunction,
'voltagesquared': AngularFrequencyResponseDomainVoltagesquared,
'currentsquared': AngularFrequencyResponseDomainCurrentsquared,
'admittancesquared': AngularFrequencyResponseDomainAdmittancesquared,
'impedancesquared': AngularFrequencyResponseDomainImpedancesquared,
'power': AngularFrequencyResponseDomainPower,
}
,'phasor': {'undefined': PhasorDomainExpression,
'voltage': PhasorDomainVoltage,
'current': PhasorDomainCurrent,
'admittance': PhasorDomainAdmittance,
'impedance': PhasorDomainImpedance,
'transfer': PhasorDomainTransferFunction,
'voltagesquared': PhasorDomainVoltagesquared,
'currentsquared': PhasorDomainCurrentsquared,
'admittancesquared': PhasorDomainAdmittancesquared,
'impedancesquared': PhasorDomainImpedancesquared,
'power': PhasorDomainPower,
}
,'phasor ratio': {'undefined': PhasorRatioDomainExpression,
'voltage': PhasorRatioDomainVoltage,
'current': PhasorRatioDomainCurrent,
'admittance': PhasorRatioDomainAdmittance,
'impedance': PhasorRatioDomainImpedance,
'transfer': PhasorRatioDomainTransferFunction,
'voltagesquared': PhasorRatioDomainVoltagesquared,
'currentsquared': PhasorRatioDomainCurrentsquared,
'admittancesquared': PhasorRatioDomainAdmittancesquared,
'impedancesquared': PhasorRatioDomainImpedancesquared,
'power': PhasorRatioDomainPower,
}
,'fourier noise': {'undefined': FourierNoiseDomainExpression,
'voltage': FourierNoiseDomainVoltage,
'current': FourierNoiseDomainCurrent,
'admittance': FourierNoiseDomainAdmittance,
'impedance': FourierNoiseDomainImpedance,
'transfer': FourierNoiseDomainTransferFunction,
'voltagesquared': FourierNoiseDomainVoltagesquared,
'currentsquared': FourierNoiseDomainCurrentsquared,
'admittancesquared': FourierNoiseDomainAdmittancesquared,
'impedancesquared': FourierNoiseDomainImpedancesquared,
'power': FourierNoiseDomainPower,
}
,'angular fourier noise': {'undefined': AngularFourierNoiseDomainExpression,
'voltage': AngularFourierNoiseDomainVoltage,
'current': AngularFourierNoiseDomainCurrent,
'admittance': AngularFourierNoiseDomainAdmittance,
'impedance': AngularFourierNoiseDomainImpedance,
'transfer': AngularFourierNoiseDomainTransferFunction,
'voltagesquared': AngularFourierNoiseDomainVoltagesquared,
'currentsquared': AngularFourierNoiseDomainCurrentsquared,
'admittancesquared': AngularFourierNoiseDomainAdmittancesquared,
'impedancesquared': AngularFourierNoiseDomainImpedancesquared,
'power': AngularFourierNoiseDomainPower,
}
,'discrete time': {'undefined': DiscreteTimeDomainExpression,
'voltage': DiscreteTimeDomainVoltage,
'current': DiscreteTimeDomainCurrent,
'admittance': DiscreteTimeDomainAdmittance,
'impedance': DiscreteTimeDomainImpedance,
'transfer': DiscreteTimeDomainTransferFunction,
'voltagesquared': DiscreteTimeDomainVoltagesquared,
'currentsquared': DiscreteTimeDomainCurrentsquared,
'admittancesquared': DiscreteTimeDomainAdmittancesquared,
'impedancesquared': DiscreteTimeDomainImpedancesquared,
'power': DiscreteTimeDomainPower,
}
,'discrete fourier': {'undefined': DiscreteFourierDomainExpression,
'voltage': DiscreteFourierDomainVoltage,
'current': DiscreteFourierDomainCurrent,
'admittance': DiscreteFourierDomainAdmittance,
'impedance': DiscreteFourierDomainImpedance,
'transfer': DiscreteFourierDomainTransferFunction,
'voltagesquared': DiscreteFourierDomainVoltagesquared,
'currentsquared': DiscreteFourierDomainCurrentsquared,
'admittancesquared': DiscreteFourierDomainAdmittancesquared,
'impedancesquared': DiscreteFourierDomainImpedancesquared,
'power': DiscreteFourierDomainPower,
}
,'Z': {'undefined': ZDomainExpression,
'voltage': ZDomainVoltage,
'current': ZDomainCurrent,
'admittance': ZDomainAdmittance,
'impedance': ZDomainImpedance,
'transfer': ZDomainTransferFunction,
'voltagesquared': ZDomainVoltagesquared,
'currentsquared': ZDomainCurrentsquared,
'admittancesquared': ZDomainAdmittancesquared,
'impedancesquared': ZDomainImpedancesquared,
'power': ZDomainPower,
}
,}




from lcapy import (symbol, nexpr, n, k, rect, sinc, sign, sin, cos, exp,
                   dt, j, pi, delta, H)

alpha = symbol('alpha')
m = symbol('m', integer=True)
N = symbol('N', integer=True, positive=True)
f0 = symbol('f0')
w0 = 2 * pi * f0

sigs = [nexpr('x(n)'), nexpr('x(a * n)'), nexpr('x(n - m)'),
        nexpr(1), delta(n), delta(n - m),
        H(n), n * H(n),
        alpha**-n * H(n),
        exp(j * 2 * pi * n / N),
        cos(2 * pi * n / N),
        sin(2 * pi * n / N)]

sigs2 = [n, n**2, 1 / n, rect(n), sinc(n), alpha**-abs(n), sign(n),
         cos(w0 * n * dt), sin(w0 * n * dt), exp(j * w0 * n * dt)]

for sig in sigs:
    print(':math:`%s \\longleftrightarrow %s`\n' %
          (sig.latex(), sig(k, N=N).simplify().latex()))



"""This module provides the Quantity class, for tracking of an expression quantity.

Copyright 2020--2021 Michael Hayes, UCECE

"""


class Quantity(object):
    """This is the base class for VoltageMixin, etc.  An Lcapy quantity is
    not a true quantity but a collection of related quantities, for
    example, voltage (with units V) and voltage spectral density (with units V/Hz)."""

    quantity = ''
    is_power = False
    is_voltage = False
    is_current = False
    is_impedance = False
    is_admittance = False
    is_transfer = False
    is_immittance = False
    is_always_causal = False
    is_undefined = False
    is_voltagesquared = False
    is_currentsquared = False
    is_impedancesquared = False
    is_admittancesquared = False
    is_ratio = False
    is_signal = False
    is_squared = False
    part = ''


class UndefinedQuantity(Quantity):

    quantity = 'undefined'
    is_undefined = True



"""This module provides the AngularFrequencyResponseDomainExpression class to
represent jomega-domain (angular frequency frequency domain) expressions.

Copyright 2022 Michael Hayes, UCECE

"""

from __future__ import division
from .domains import AngularFrequencyResponseDomain
from .inverse_fourier import inverse_fourier_transform
from .expr import Expr, expr, expr_make
from .state import state, validate
from .sym import fsym, ssym, tsym, omegasym, j, pi
from .units import u as uu
from sympy import Expr as symExpr

__all__ = ('AngularFrequencyResponseDomainExpression', )


class AngularFrequencyResponseDomainExpression(AngularFrequencyResponseDomain, Expr):
    """Frequency domain expression or symbol (angular frequency)."""

    var = omegasym

    def __init__(self, val, **assumptions):

        check = assumptions.pop('check', True)

        super(AngularFrequencyResponseDomainExpression,
              self).__init__(val, **assumptions)

        expr = self.expr

        if check and not expr.has(omegasym):
            if expr.has(tsym):
                validate(state.t_in_jw,
                         'jomega-domain expression %s depends on t' % expr)
            if expr.has(ssym):
                validate(state.s_in_jw,
                         'jomega-domain expression % s depends on s' % expr)
            if expr.has(omegasym):
                validate(state.f_in_jw,
                         'jomega-domain expression %s depends on f' % expr)

    def _div_compatible_domains(self, x):

        if x.is_angular_fourier_domain:
            return True

        return super(AngularFrequencyResponseDomainExpression, self)._div_compatible_domains(x)

    def as_expr(self):
        return AngularFrequencyResponseDomainExpression(self)

    def inverse_fourier(self, **assumptions):
        """Attempt inverse Fourier transform."""

        expr = self.subs(2 * pi * fsym)
        result = inverse_fourier_transform(expr.sympy, fsym, tsym)

        return self.change(result, 'time', units_scale=uu.Hz, **assumptions)

    def time(self, **assumptions):
        """Alias for inverse_fourier."""

        return self.inverse_fourier(**assumptions)

    def angular_fourier(self, **assumptions):
        """Convert to angular Fourier domain."""

        return self.laplace(**assumptions).angular_fourier()

    def fourier(self, **assumptions):
        """Convert to Fourier domain."""

        return self.laplace(**assumptions).fourier()

    def norm_fourier(self, **assumptions):
        """Convert to normalized Fourier domain."""

        return self.laplace(**assumptions).norm_fourier()

    def norm_angular_fourier(self, **assumptions):
        """Convert to normalized angular Fourier domain."""

        return self.laplace(**assumptions).norm_angular_fourier()

    def laplace(self, **assumptions):
        """Convert to Laplace domain."""

        from .symbols import s, j

        return self.subs(s / j)

    def phasor(self, **assumptions):
        """Convert to phasor domain."""

        return self.time(**assumptions).phasor(**assumptions)

    def plot(self, wvector=None, **kwargs):
        """Plot angular frequency response at values specified by wvector.

        There are many plotting options, see matplotlib.pyplot.plot.

        For example:
            V.plot(fvector, log_frequency=True)
            V.real.plot(fvector, color='black')
            V.phase.plot(fvector, color='black', linestyle='--')

        By default complex data is plotted as separate plots of magnitude (dB)
        and phase.

        `kwargs` include:
        `axes` - the plot axes to use otherwise a new figure is created
        `xlabel` - the x-axis label
        `ylabel` - the y-axis label
        `ylabel2` - the second y-axis label if needed, say for mag and phase
        `xscale` - the x-axis scaling, say for plotting as ms
        `yscale` - the y-axis scaling, say for plotting mV
        `norm` - use normalized frequency
        `dbmin` - the smallest value to plot in dB (default -120)
        in addition to those supported by the matplotlib plot command.

        The plot axes are returned.  This is a tuple for magnitude/phase or
        real/imaginary plots.
        """

        from .plot import plot_angular_frequency
        return plot_angular_frequency(self, wvector, **kwargs)

    def bode_plot(self, wvector=None, unwrap=True, **kwargs):
        """Plot frequency response for a frequency-domain phasor as a Bode
        plot (but without the straight line approximations).  wvector
        specifies the angular frequencies.  If it is a tuple (f1, f2), it sets
        the frequency limits.   Since a logarithmic frequency scale is used,
        f1 must be greater than 0.

        `unwrap` controls phase unwrapping (default True).

        For more info, see `plot`.
        """

        from .plot import plot_angular_bode
        return plot_angular_bode(self, wvector, unwrap=unwrap, **kwargs)

    def nyquist_plot(self, wvector=None, log_frequency=True, **kwargs):
        """Plot frequency response as a Nyquist plot (imaginary part versus
        real part).  wvector specifies the angular frequencies.  If it
        is a tuple (f1, f2), it sets the frequency limits as (f1, f2).

        `npoints` set the number of plotted points.

        The unit circle is shown by default.  This can be disabled
        with `unitcircle=False`.

        """

        from .plot import plot_nyquist
        return plot_nyquist(self, wvector, log_frequency=log_frequency, **kwargs)

    def nichols_plot(self, wvector=None, log_frequency=True, **kwargs):
        """Plot frequency response as a Nichols plot (dB versus phase).
        wvector specifies the angular frequencies.  If it is a tuple
        (f1, f2), it sets the frequency limits as (f1, f2).

        `npoints` set the number of plotted points.

        """

        from .plot import plot_nichols
        return plot_nichols(self, wvector, log_frequency=log_frequency, **kwargs)


jomega = AngularFrequencyResponseDomainExpression('j * omega')
jomega.units = uu.rad / uu.s



"""This module provides the AngularFourierNoiseDomainExpression class to represent omega-domain (angular Fourier domain) noise expressions.

Copyright 2014--2022 Michael Hayes, UCECE
"""

from __future__ import division
from .sym import symsimplify
from .functions import sqrt
from .sym import pi, omegasym, fsym
from .state import state
from .domains import AngularFourierNoiseDomain
from .expr import expr
from .noiseexpr import NoiseExpression
from .fexpr import f, FourierDomainExpression
from .omegaexpr import AngularFourierDomainExpression
import sympy as sym


class AngularFourierNoiseDomainExpression(AngularFourierNoiseDomain, NoiseExpression):
    """Angular frequency domain (one-sided) noise spectrum expression (amplitude
    spectral density).

    This characterises a wide-sense stationary, zero-mean Gaussian
    noise random process.

    When performing arithmetic on two AngularFourierNoiseDomainExpression
    expressions it is assumed that they are uncorrelated unless they have
    the same nid (noise indentifier).  If the nid is not specified, a new one is created.

    Uncorrelated noise expressions are added in quadrature (on a power
    basis).  Thus (AngularFourierNoiseDomainExpression(3) +
    AngularFourierNoiseDomainExpression(4)).expr = 5 since 5 =
    sqrt(3**2 + 4**2)

    AngularFourierNoiseDomainExpression(3) !=
    AngularFourierNoiseDomainExpression(3) since they are different
    noise realisations albeit with the same properties.  However,
    AngularFourierNoiseDomainExpression(3).expr ==
    AngularFourierNoiseDomainExpression(3).expr.  Similarly,
    AngularFourierNoiseDomainExpression(3, nid='n1') ==
    AngularFourierNoiseDomainExpression(3, nid='n1') since they have
    the same noise identifier and thus have the same realisation.

    Caution: The sum of two noise expressions generates a noise
    expression with a new nid.  This can lead to unexpected results
    since noise expressions with different nids are assumed to be
    uncorrelated.  For example, consider:
    a = AngularFourierNoiseDomainExpression(3);
    b = AngularFourierNoiseDomainExpression(4)
    a + b - b gives sqrt(41) and a + b - a gives sqrt(34).

    This case is correctly handled by the SuperpositionVoltage and
    SuperpositionCurrent classes since each noise component is stored
    and considered separately.

    (SuperpositionVoltage(a) + SuperpositionVoltage(b) - SuperpositionVoltage(b)).n gives 3 as expected.

    """
    var = omegasym

    def plot(self, omegavector=None, **kwargs):
        """Plot frequency response at values specified by omegavector.

        There are many plotting options, see matplotlib.pyplot.plot.

        For example:
            V.plot(omegavector, log_frequency=True)
            V.real.plot(omegavector, color='black')
            V.phase.plot(omegavector, color='black', linestyle='--')

        By default complex data is plotted as separate plots of magnitude (dB)
        and phase.
        """
        from .plot import plot_angular_frequency

        return plot_angular_frequency(self, omegavector, **kwargs)

    def transform(self, arg, **assumptions):
        """Transform into a different domain."""

        arg = expr(arg)
        if isinstance(arg, FourierDomainExpression):
            result = self.subs(f * 2 * pi)
            cls = self._class_by_quantity(self.quantity, 'fourier noise')
            return cls(result, nid=self.nid, **assumptions)
        elif isinstance(arg, AngularFourierDomainExpression):
            result = self.subs(arg, **assumptions)
            cls = self._class_by_quantity(
                self.quantity, 'angular fourier noise')
            return cls(result, nid=self.nid, **assumptions)

        return super(AngularFourierNoiseDomainExpression, self).transform(arg, **assumptions)


class AngularFourierNoiseDomainVoltage(AngularFourierNoiseDomainExpression):
    """Voltage noise amplitude spectral density (units V/rtrad/s).
    This can be a function of angular frequency, omega.  For example,
    to model an opamp voltage noise:

    v = AngularFourierNoiseDomainVoltage(1e-8 / sqrt(omega) + 8e-9)

    """

    quantity_label = 'Voltage noise spectral density'
    units = 'V/rtrad/s'


class AngularFourierNoiseDomainCurrent(AngularFourierNoiseDomainExpression):
    """Current noise amplitude spectral density (units A/rtrad/s).

    This can be a function of angular frequency, omega.  For example,
    to model an opamp current noise:

    i = AngularFourierNoiseDomainCurrent(3e-12 / sqrt(omega) + 200e-15)
    """

    quantity_label = 'Current noise spectral density'
    units = 'A/rtrad/s'


from .omegaexpr import omega  # nopep8
from .noisefexpr import FourierNoiseDomainExpression  # nopep8



"""This module wraps system dependent programs for pdf generation, etc.

Copyright 2014--2021 Michael Hayes, UCECE

"""


from os import path, remove, chdir, getcwd, stat, system
from subprocess import call
import re
import platform

try:
    from subprocess import DEVNULL
except ImportError:
    # For python 2.7
    import os
    DEVNULL = open(os.devnull, 'wb')


# System dependent functions

# Note, sometime in 2018 ImageMagick's convert program by default
# disallowed pdf file conversions.  A work-around is to edit
# /etc/ImageMagick-6/policy.xml to enable this conversion.

programs = {}


def which(program):

    import os

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    for path1 in os.environ["PATH"].split(os.pathsep):
        exe_file = os.path.join(path1, program)
        if is_exe(exe_file):
            return exe_file

    return None


def hasexe(program):

    try:
        return programs[program]
    except:
        if which(program) is not None:
            programs[program] = True
        else:
            programs[program] = which(program + '.exe') is not None

    return programs[program]


def checkexe(command, debug=False):

    if isinstance(command, list):
        program = command[0]
        command = ' '.join(command)
    else:
        program = command.split(' ')[0]

    if hasexe(program):
        if debug:
            print('Checking: %s found' % program)
    else:
        if debug:
            print('Checking: %s not found' % program)
        raise RuntimeError('%s is not installed' % program)


def run(command, stderr=DEVNULL, stdout=DEVNULL, shell=False, debug=False):

    checkexe(command, debug=debug)
    if debug:
        print('Running: %s' % ' '.join(command))
    call(command, stderr=stderr, stdout=stdout, shell=shell)


def tmpfilename(suffix='', dirname=None):

    from tempfile import gettempdir, NamedTemporaryFile

    if dirname is None:
        # Searches using TMPDIR, TEMP, TMP environment variables
        dirname = gettempdir()

    filename = NamedTemporaryFile(suffix=suffix, dir=dirname,
                                  delete=False).name
    return filename


class PDFConverter(object):

    def __init__(self, debug=False):

        self.debug = debug

    def to_svg(self, pdf_filename, svg_filename):

        run(['pdf2svg', pdf_filename, svg_filename], debug=self.debug)
        if not path.exists(svg_filename):
            raise RuntimeError('Could not generate %s with pdf2svg.  Is it installed?' %
                               svg_filename)

    def to_png_convert(self, pdf_filename, png_filename, dpi=300):

        program = 'convert'
        if platform.system() == 'Windows':
            program = 'magick convert'

        run([program, '-density %d' %
            int(dpi), pdf_filename, png_filename], debug=self.debug)

        if not path.exists(png_filename):
            raise RuntimeError('Could not generate %s with convert' %
                               png_filename)
        if stat(png_filename).st_size == 0:
            raise RuntimeError('Could not generate %s with convert, empty file' %
                               png_filename)

    def to_png_ghostscript(self, pdf_filename, png_filename, dpi=300):

        program = 'gs'
        if platform.system() == 'Windows':
            program = 'gswin32'
            if platform.machine().endswith('64'):
                program = 'gswin64'

        run([program, '-q', '-dQUIET', '-dSAFER', '-dBATCH', '-dNOPAUSE',
             '-dNOPROMPT',  '-dMaxBitmap=500000000',  '-dAlignToPixels=0',
             '-dGridFitTT=2',  '-sDEVICE=pngalpha',  '-dTextAlphaBits=4',
             '-dGraphicsAlphaBits=4',  '-r%dx%d' % (int(dpi), int(dpi)),
             '-sOutputFile=' + png_filename, pdf_filename], debug=self.debug)

    def to_png_pdftoppm(self, pdf_filename, png_filename, dpi=300):

        root, ext = path.splitext(png_filename)
        args = ['pdftoppm', '-r %d' %
                int(dpi), '-png', '-thinlinemode shape', '-singlefile', pdf_filename, root]
        if False:
            # TODO, determine why this fails...
            run(args, debug=self.debug)
        else:
            checkexe(args)
            system(' '.join(args))

        if not path.exists(png_filename):
            raise RuntimeError('Could not generate %s with pdftoppm' %
                               png_filename)

    def to_png(self, pdf_filename, png_filename, dpi=300, method=None):

        converters = {'ghostscript': self.to_png_ghostscript,
                      'convert': self.to_png_convert,
                      'pdftoppm': self.to_png_pdftoppm}

        # Ordered in preference
        methods = ('ghostscript', 'convert', 'pdftoppm')

        if method is not None:
            if method not in methods:
                raise KeyError('Unknown method %s, known methods: %s' %
                               (method, ', '.join(methods)))

            # Try the specified method then fall back to default methods
            # if it does not work
            func = converters[method]
            return func(pdf_filename, png_filename, dpi)

        for method in methods:
            func = converters[method]
            try:
                return func(pdf_filename, png_filename, dpi)
            except:
                pass

        raise RuntimeError(
            """Could not convert pdf to png, tried: %s.  Check that one of these programs is installed.  See https://lcapy.readthedocs.io/en/latest/install.html""" % ', '.join(methods))


def run_dot(dotfilename, filename):

    base, ext = path.splitext(filename)
    #run([which('dot'), '-T ' + ext[1:], '-o ' + filename, dotfilename])
    system('dot -T %s -o %s %s' % (ext[1:], filename, dotfilename))
    remove(dotfilename)
    return


class LatexRunner(object):

    def __init__(self, debug=False):

        self.debug = debug

    def run(self, tex_filename):

        checkexe('pdflatex', debug=self.debug)

        root, ext = path.splitext(tex_filename)
        dirname = path.dirname(tex_filename)
        baseroot = path.basename(root)
        cwd = getcwd()
        if dirname != '':
            if self.debug:
                print('Chdir: %s' % dirname)
            chdir(path.abspath(dirname))

        run(['pdflatex', '-interaction', 'batchmode',
            baseroot + '.tex'], debug=self.debug)

        if dirname != '':
            if self.debug:
                print('Chdir: %s' % cwd)
            chdir(cwd)

        return root + '.pdf'

    def cleanup(self, tex_filename, wanted_filename=''):

        root, ext = path.splitext(tex_filename)
        exts = ['.tex', '.aux', '.log', '.pdf']
        filenames = [root + ext for ext in exts]

        if wanted_filename in filenames:
            filenames.remove(wanted_filename)

        for filename in filenames:
            if path.exists(filename):
                if self.debug:
                    print('Removing: %s' % filename)
                remove(filename)

    def extract_circuitikz_version(self, tex_filename):

        log_filename = tex_filename.replace('.tex', '.log')

        with open(log_filename, 'rt') as logfile:
            lines = logfile.readlines()

        date = None
        version = None
        for m, line in enumerate(lines):
            if line.startswith('Package: circuitikzgit'):
                match = re.search(r'circuitikzgit ([0-9/]+)', line)
                date = match.group(1)
                # There is an annoying newline in the middle of the word version
                version = lines[m + 1].strip('ion').strip()
                break
            if line.startswith('Package: circuitikz'):
                match = re.search(r'circuitikz ([0-9/]+)', line)
                date = match.group(1)
                version = lines[m + 1].strip()
                break

        return date, version

    def find_circuitikz_version(self):

        content = r"""
        \documentclass[a4paper]{standalone}
        \usepackage{circuitikz}
        \begin{document}
        \end{document}"""

        tex_filename = tmpfilename('.tex')
        open(tex_filename, 'w').write(content)

        self.run(tex_filename)

        date, version = self.extract_circuitikz_version(tex_filename)

        self.cleanup(tex_filename)
        return date, version



from lcapy import *
from lcapy.fexpr import FourierDomainExpression
from lcapy.expr import TimeDomainExpression
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_fourier(self):

        self.assertEqual(DiracDelta(t).FT(), 1, "DiracDelta(t)")
        self.assertEqual(TimeDomainExpression('x(t)').FT(),
                         FourierDomainExpression('X(f)'), "x(t)")
        self.assertEqual(TimeDomainExpression('5 * x(t)').FT(),
                         FourierDomainExpression('5 * X(f)'), "5 * x(t)")

    def test_inverse_fourier(self):

        self.assertEqual((f * 0 + 1).IFT(), DiracDelta(t),
                         "1")
        self.assertEqual((f * 0 + 10).IFT(), 10
                         * DiracDelta(t), "0")
        self.assertEqual(FourierDomainExpression('V(f)').IFT(),
                         TimeDomainExpression('v(t)'), "V(f)")
        self.assertEqual(FourierDomainExpression('10 * V(f)').IFT(),
                         TimeDomainExpression('10 * v(t)'), "V(f)")
        self.assertEqual(TimeDomainExpression('v(t)').FT().IFT(),
                         TimeDomainExpression('v(t)'), "v(t)")
        self.assertEqual(TimeDomainExpression('v(t/2)').FT().IFT(),
                         TimeDomainExpression('v(t/2)'), "v(t/2)")

    def test_fourier2(self):

        # 1
        self.assertEqual((t * 0 + 1).FT(), DiracDelta(f))
        self.assertEqual((t * 0 + 1).FT().IFT(), 1)

        # t
        self.assertEqual(t.FT(), j * DiracDelta(f, 1) / (2 * pi))
        self.assertEqual(t.FT().IFT(), t)

        # t**2
        self.assertEqual((t**2).FT(), -DiracDelta(f, 2) / (2 * pi)**2)
        self.assertEqual((t**2).FT().IFT(), t**2)

        # 2 * cos(2 * pi * t)
        self.assertEqual(2 * cos(2 * pi * t).FT(),
                         DiracDelta(f - 1) + DiracDelta(f + 1))

        # 2 * sin(2 * pi * t)
        self.assertEqual(2 * sin(2 * pi * t).FT(),
                         -j * DiracDelta(f - 1) + j * DiracDelta(f + 1))

        # exp(j * 2 * pi * t)
        self.assertEqual(exp(j * 2 * pi * t).FT(), DiracDelta(f - 1))
        self.assertEqual(exp(j * 2 * pi * t).FT().IFT(),
                         exp(j * 2 * pi * t))

        # exp(-j * 2 * pi * t)
        self.assertEqual(exp(-j * 2 * pi * t).FT(), DiracDelta(f + 1))
        self.assertEqual(exp(-j * 2 * pi * t).FT().IFT(),
                         exp(-j * 2 * pi * t))

        # u(t)
        self.assertEqual(u(t).FT(), DiracDelta(f) / 2 + 1 / (j * 2 * pi * f))
        # TODO teach SymPy that sign(t) + 1 = 2 * u(t)
        # self.assertEqual(u(t).FT().IFT(), u(t))

        # t * u(t)
        self.assertEqual((t * u(t)).FT(), -1 / (2 * pi * f) **
                         2 + j * DiracDelta(f, 1) / (4 * pi))
        # TODO teach SymPy that sign(t) + 1 = 2 * u(t)
        # self.assertEqual((t * u(t)).FT().IFT(), t * u(t))

        # sign(t)
        self.assertEqual(sign(t).FT(), 1 / (j * pi * f))
        self.assertEqual(sign(t).FT().IFT(), sign(t))

        # t * sign(t)
        self.assertEqual((t * sign(t)).FT(), -1 / (2 * pi**2 * f**2))
        self.assertEqual((t * sign(t)).FT().IFT(), t * sign(t))

        # 2 * cos(2 * pi * t + 3)
        self.assertEqual(2 * cos(2 * pi * t + 3).FT(),
                         DiracDelta(f - 1) * exp(3 * j) +
                         DiracDelta(f + 1) * exp(-3 * j))

        self.assertEqual(2 * rect(t).FT(), 2 * sinc(f))
        self.assertEqual(rect(t / 2).FT(), 2 * sinc(2 * f))

        self.assertEqual(2 * sinc(t).FT(), 2 * rect(f))
        self.assertEqual(sinc(2 * t).FT(), rect(f / 2) / 2)

        self.assertEqual((rect(t) * cos(2 * pi * t)).FT(),
                         sinc(f - 1) / 2 + sinc(f + 1) / 2)

        self.assertEqual(tri(t).FT(), sinc(f)**2)

        self.assertEqual(trap(t, 1).FT(), sinc(f)**2)
        self.assertEqual(trap(t, 0).FT(), sinc(f))
        self.assertEqual(trap(t, 0.5).FT(), 0.5 * sinc(0.5 * f) * sinc(f))

        # Time shift
        # rect(t - 2)
        self.assertEqual(rect(t - 2).FT(), sinc(f) * exp(-j * 4 * pi * f))
        self.assertEqual(rect(t - 2).FT().IFT(), rect(t - 2))

        # Frequency shift
        # rect(t) * exp(j * 4 * pi * t)
        self.assertEqual((rect(t) * exp(j * 4 * pi * t)).FT(), sinc(f - 2))
        self.assertEqual((rect(t) * exp(j * 4 * pi * t)).FT().IFT(),
                         (rect(t) * exp(j * 4 * pi * t)))

    def test_fourier_convolution(self):

        a = expr('Integral(3 * x(t - tau) * y(tau), (tau, -oo, oo))')
        A = a(f)
        self.assertEqual(A, expr('3 * X(f) * Y(f)'), "3 * X * Y")
        # self.assertEqual(A(t), a, "a(f)(t)")

        a = expr('Integral(3 * x(tau) * y(t - tau), (tau, -oo, oo))')
        A = a(f)
        self.assertEqual(A, expr('3 * X(f) * Y(f)'), "3 * X * Y")
        # self.assertEqual(A(t), a, "a(f)(t)")



from .quantity import Quantity


class CurrentMixin(Quantity):

    quantity = 'current'
    quantity_label = 'Current'
    quantity_units = 'A'
    is_current = True
    is_signal = True
    is_undefined = False

    def cpt(self):
        from .oneport import I, Idc, Iac

        if self.is_dc:
            return Idc(self.expr)
        elif self.is_phasor_domain:
            return Iac(self, omega=self.omega)

        return I(self.expr)



from .domains import domains
from .quantities import quantities
from .seqclasses import seqclasses


def seqmap(quantity, domain):

    domaincls = domains[domain]

    if quantity == 'undefined':
        return seqclasses[domain][quantity]

    return seqclasses[domain][quantity]



from matplotlib.pyplot import savefig
from lcapy import Circuit, j, f, s, degrees, pi

b = Circuit('opamp-voltage-follower-RC-load-open-loop.sch')
G = b.transfer(2, 0, 4, 0)

f1 = 10
f2 = 1e6
A0 = 1e6

A = A0 * (1 / (1 + j * f / f1)) * (1 / (1 + j * f / f2))
A = A.subs(f, s / (j * 2 * pi))

Ro = 40
R = 20
C = 100e-9

G = G.subs({'A': A, 'Ro': Ro, 'R': R, 'C': C})

ax = G.bode_plot((1, 10e6), plot_type='dB-degrees')
ax[0].set_ylim(-30, 130)
ax[1].set_ylim(-240, 0)

savefig(__file__.replace('.py', '.png'), bbox_inches='tight')



"""This module provides node support for schematic drawing.

Copyright 2022--2023 Michael Hayes, UCECE

"""

from .latex import latex_format_node_label
from .opts import Opts


class Node:

    def __init__(self, name):

        self.name = name
        self._port = False
        self._count = 0
        parts = name.split('_')
        # Primary 1, 2, a, a_3.  Not primary _1, _2, _a, _a_3, 0_3, a_b_c
        self.primary = (name[0] != '_' and len(parts) <= 2) and not (
            name[0].isdigit() and len(parts) != 1)

        # See MX1.sch for example.
        if '._' in name:
            self.primary = False

        self.elt_list = []
        self.pos = 'unknown'
        # Sanitised name
        self.s = name.replace('.', '@')
        self.label = latex_format_node_label(self.name)
        self.pin = False
        self.pinlabel = ''
        self.namepos = None
        self.pinname = ''
        self.pinpos = None
        self.clock = False
        self.auxiliary = None
        # Node has an implicit connection, say to ground.
        self.implicit = False
        # Implicit symbol to use if node is an autoground
        self._implicit_symbol = None
        self.split_count = 0
        self.opts = Opts()
        self.drawn = False
        self.label_drawn = False
        self.parent = None
        self.is_split = False

    @property
    def basename(self):
        fields = self.name.split('.')
        return fields[-1]

    @property
    def cptname(self):
        fields = self.name.split('.')
        if len(fields) < 2:
            return None
        return fields[-2]

    def __repr__(self):
        return '%s @ (%s)' % (self.name, self.pos)

    def append(self, elt):
        """Add new element to the node"""

        if elt.type == 'P':
            self._port = True

        self.elt_list.append(elt)
        if elt.type not in ('A', 'O'):
            self._count += 1

    @property
    def count(self):
        """Number of elements (including wires but not open-circuits and
        annotations) connected to the node"""

        if self.implicit:
            return self._count + 1
        return self._count

    def belongs(self, cpt_name):
        return self.cptname == cpt_name

    def visible(self, draw_nodes):
        """Return True if node drawn.
        `draw_nodes' can be `all', 'none', 'connections', 'primary', None,
        True, or False."""

        if self.auxiliary:
            return False

        if draw_nodes in ('all', True):
            return True

        if self.pin:
            return False

        if self._port:
            return True

        if draw_nodes in ('none', None, False):
            return False

        # Implied port
        if self.count == 1:
            return True

        if draw_nodes in ('connections', 'connected'):
            return self.count > 2

        if draw_nodes == 'primary':
            return self.primary

        raise ValueError('Unknown argument %s for draw_nodes' % draw_nodes)

    @property
    def is_port(self):
        """Return True if node is a port"""

        return self._port

    @property
    def is_dangling(self):
        """Return True if node has a single connection"""

        return self.count == 1

    @property
    def is_ground(self):
        """Return True if node is a ground"""

        return self.name.startswith('0')

    def show_label(self, label_nodes):

        if self.label == '':
            return False

        name = self.basename

        # pins is for backward compatibility
        if label_nodes in ('none', 'pins', 'false', False):
            return False
        elif label_nodes in ('all', 'true', True):
            return True
        elif label_nodes == 'alpha':
            return self.primary and name[0].isalpha()
        elif label_nodes == 'primary':
            return self.primary

        # handle label_nodes = '{1, 2}' etc.
        if label_nodes[0] == '{' and label_nodes[-1] == '}':
            label_nodes = label_nodes[1:-1]
        labels = [foo.strip() for foo in label_nodes.split(',')]
        return self.label in labels

    def debug(self):
        print(' %s @ (%s), count=%d, pin=%s' % (self.name, self.pos,
                                                self._count, self.pin))

    def split(self, elt):

        if self.count == 1:
            return self

        self.split_count += 1
        name = self.name + '_split%d' % (self.split_count - 1)

        new_node = Node(name)
        self._count -= 1
        new_node.parent = self
        new_node._count = 1
        new_node._port = self._port
        new_node.label = self.label
        new_node._port = elt.type == 'P'
        new_node.opts = self.opts.copy()
        new_node.is_split = True

        return new_node

    @property
    def implicit_symbol(self):
        """Return node implicit_symbol"""

        return self._implicit_symbol

    @implicit_symbol.setter
    def implicit_symbol(self, value):
        """Set node implicit_symbol."""

        if self._implicit_symbol != None and self._implicit_symbol != value:
            raise ValueError('Conflicting symbols %s and %s' %
                             (self._implicit_symbol, value))
        self._implicit_symbol = value



"""
Copyright 2014--2022 Michael Hayes, UCECE
"""

from __future__ import division
from .expr import expr
from .sexpr import s
from .printing import latex, pretty
from .circuit import Circuit
from .state import state
from warnings import warn


class Network(object):
    """This is the base class for network objects."""

    is_voltage_source = False
    is_current_source = False
    is_inductor = False
    is_capacitor = False
    is_resistor = False
    is_conductor = False
    is_parallel = False
    is_series = False
    is_noiseless = True

    # True if initial conditions are zero (or unspecified).
    zeroic = True

    # None if component does not have initial conditions.
    # True if initial conditions are specified.
    # False if initial conditions are not specified.
    has_ic = None

    cpt_type = ''
    netkeyword = ''
    kwargs = {}

    def _opts_str(self, arg=None, **kwargs):

        parts = []
        if arg is not None and arg != '':
            parts.append(arg)

        for key, val in kwargs.items():
            parts.append('%s=%s' % (key, val))

        for key, val in self.kwargs.items():
            parts.append('%s=%s' % (key, val))
        return ', '.join(parts)

    def __repr__(self):

        parts = []
        for key, val in self.kwargs.items():
            parts.append("%s='%s'" % (key, val))

        argsrepr = ', '.join([repr(arg) for arg in self.args] + parts)
        return '%s(%s)' % (self.__class__.__name__, argsrepr)

    def __str__(self):

        argsrepr = ', '.join([repr(arg) for arg in self.args])
        return '%s(%s)' % (self.__class__.__name__, argsrepr)

    def _repr_pretty_(self, p, cycle):

        p.text(self.pretty())

    def _repr_latex_(self):

        s = self.latex(mode='plain')
        return "$$%s$$" % s

    def _repr_png_(self):

        return None

    def _repr_svg_(self):

        return None

    def pretty(self, **kwargs):

        def pretty1(arg, **kwargs):
            try:
                return arg.pretty(**kwargs)
            except:
                return pretty(arg, **kwargs)

        parts = []
        for key, val in self.kwargs.items():
            parts.append("%s='%s'" % (key, val))

        argsrepr = ', '.join([pretty1(arg, **kwargs)
                             for arg in self.args] + parts)
        return '%s(%s)' % (self.__class__.__name__, argsrepr)

    def latex(self, **kwargs):

        def latex1(arg, **kwargs):
            try:
                return arg.latex(**kwargs)
            except:
                return latex(arg, **kwargs)

        argsrepr = ', '.join([latex1(arg, **kwargs) for arg in self.args])
        return '\\mathrm{%s}(%s)' % (self.__class__.__name__, argsrepr)

    def pprint(self):

        print(self.pretty())

    @property
    def analysis(self):
        return self.cct.analysis

    def describe(self):
        """Print a message describing how network is solved."""
        return self.cct.describe()

    def simplify(self):

        return self

    def _add_elements(self):

        netlist = self.netlist()
        for net in netlist.split('\n'):
            self._add(net)

        # Hack, create ground reference.
        self._add('W %d 0' % (self._node - 1))

    def _net_make(self, netlistmaker, n1=None, n2=None, dir='right'):

        net = self
        if n2 == None:
            n2 = net._node
        if n1 == None:
            n1 = net._node

        cpt_type = net.__class__.__name__ if net.cpt_type == '' else net.cpt_type

        opts_str = self._opts_str(dir)
        name = netlistmaker._make_name(cpt_type, self.args)
        if net.netkeyword != '':
            return '%s %s %s %s %s; %s' % (name, n1, n2, net.netkeyword,
                                           netlistmaker._netargs(net), opts_str)
        else:
            return '%s %s %s %s; %s' % (name, n1, n2,
                                        netlistmaker._netargs(net), opts_str)

    @property
    def _depths(self):
        return [net._depth for net in self.args]

    @property
    def _depth(self):
        from .oneport import Ser, Par

        if not isinstance(self, (Ser, Par)):
            return 0

        depths = self._depths
        return 1 + max(depths)

    def netlist(self, layout='horizontal', evalf=None):
        """Create a netlist.

        `layout` can be 'horizontal', 'vertical', or 'ladder'.

        `evalf` can be False or an integer specifying the number of
        decimal places used to evaluate floats.
        """

        if layout == 'ladder':
            from .laddermaker import LadderMaker
            return LadderMaker(self, layout=layout, evalf=evalf)()

        from .netlistmaker import NetlistMaker
        return NetlistMaker(self, layout=layout, evalf=evalf)()

    def pdb(self):
        """Enter the python debugger."""

        import pdb
        pdb.set_trace()
        return self

    def sch(self, layout='horizontal', evalf=False):
        """Convert a Network object into a Schematic object."""

        from .schematic import Schematic

        netlist = self.netlist(layout=layout, evalf=evalf)
        sch = Schematic()
        for net in netlist.split('\n'):
            sch.add(net)

        return sch

    def draw(self, filename=None, layout='horizontal', form=None, evalf=False, **kwargs):
        """Draw schematic of network.

        filename specifies the name of the file to produce.  If None,
        the schematic is displayed on the screen.

        Note, if using Jupyter, then need to first issue command %matplotlib inline

        `layout` is either 'horizontal', 'vertical', or 'ladder'.

        `evalf` can be False or an integer specifying the number of
        decimal places used to evaluate floats.

        `kwargs` include:
           label_ids: True to show component ids
           label_values: True to display component values
           draw_nodes: True to show all nodes, False to show no nodes,
             'primary' to show primary nodes,
             'connections' to show nodes that connect more than two components,
             'all' to show all nodes
           label_nodes: True to label all nodes, False to label no nodes,
             'primary' to label primary nodes,
             'alpha' to label nodes starting with a letter,
             'pins' to label nodes that are pins on a chip,
             'all' to label all nodes
           style: 'american', 'british', or 'european'
           scale: schematic scale factor, default 1.0
           node_spacing: spacing between component nodes, default 2.0
           cpt_size: size of a component, default 1.5
           dpi: dots per inch for png files
           help_lines: distance between lines in grid, default 0.0 (disabled)
           debug: True to display debug information
        """

        if 'label_ids' not in kwargs:
            kwargs['label_ids'] = False
        if 'label_values' not in kwargs:
            kwargs['label_values'] = True
        if 'label_nodes' not in kwargs:
            kwargs['label_nodes'] = False
        if 'draw_nodes' not in kwargs:
            kwargs['draw_nodes'] = 'connections'

        if form is not None:
            warn('Form is deprecated; use layout instead')
            layout = form

        self.sch(layout=layout, evalf=evalf).draw(filename=filename, **kwargs)

    @property
    def cct(self):
        """Convert a Network object into a Circuit object."""

        if hasattr(self, '_cct'):
            return self._cct

        netlist = self.netlist()
        cct = Circuit()
        for net in netlist.split('\n'):
            cct.add(net)
        self._cct = cct
        return cct

    def circuit(self):
        """Convert a Network object into a Circuit object."""

        return self.cct

    @property
    def initial_value_problem(self):
        return self.cct.initial_value_problem

    @property
    def is_IVP(self):
        return self.cct.is_IVP

    @property
    def is_dc(self):
        return self.cct.is_dc

    @property
    def is_ac(self):
        return self.cct.is_ac

    @property
    def is_causal(self):
        return self.cct.is_causal

    @property
    def has_dc(self):
        return self.cct.has_dc

    @property
    def has_ac(self):
        return self.cct.has_ac

    @property
    def has_transient(self):
        return self.cct.has_transient

    @property
    def kinds(self):
        """Return list of transform domain kinds."""
        return self.cct.kinds

    @property
    def params(self):
        """Return list of symbols used as arguments in the network."""

        # This would be faster to determine directly from AST.
        return self.cct.params

    @property
    def symbols(self):
        """Return dictionary of symbols defined in the network."""

        return self.cct.symbols

    def transform(self, form='cauerI'):
        """Transform the network impedance into an alternative form.  The
        transformation is performed using network synthesis of the
        network's impedance (note, this ignores the sources).  `form`
        includes: cauerI, cauerII, fosterI, fosterII.

        Note some methods generate networks with negative value
        components.

        """

        return self.Z(s).network(form)

    def subs(self, subs_dict):
        """Substitute values using dictionary of substitutions.

        For example, b = a.subs({'R1': 1e3, 'R2': 9e3})"""

        if not self.is_parallel and not self.is_series:
            return self.__class__(*[str(expr(arg).subs(subs_dict)) for arg in self.args])

        return self.__class__(*[arg.subs(subs_dict) for arg in self.args])

    def noisy(self, T='T'):
        """Create noisy network model by replacing resistances with a series
        combination of a resistance and a noise voltage source."""

        if self.is_resistor and not self.is_noiseless:
            from .oneport import Vnoise, NR

            Vn = 'sqrt(4 * k_B * %s * %s)' % (T, self.args[0])
            return NR(*self.args) + Vnoise(Vn)

        if self.is_conductor and not self.is_noiseless:
            from .oneport import Vnoise, NG

            Vn = 'sqrt(4 * k_B * %s / (%s))' % (T, self.args[0])
            return NG(*self.args) + Vnoise(Vn)

        if not self.is_parallel and not self.is_series:
            return self.__class__(*self.args)

        return self.__class__(*[arg.noisy(T=T) for arg in self.args])



from .cexpr import ConstantDomainExpression
from .fexpr import FourierDomainExpression
from .sexpr import LaplaceDomainExpression
from .texpr import TimeDomainExpression
from .phasor import PhasorDomainExpression
from .noiseomegaexpr import AngularFourierNoiseDomainExpression


classmap = {'s': LaplaceDomainExpression,
            'ivp': LaplaceDomainExpression,
            'laplace': LaplaceDomainExpression,
            't': TimeDomainExpression,
            'time': TimeDomainExpression,
            'f': FourierDomainExpression,
            'dc': ConstantDomainExpression,
            'n': AngularFourierNoiseDomainExpression}

symbolmap = {'s': '(s)',
             'laplace': '(s)',
             't': '(t)',
             'time': '(t)',
             'f': '(f)',
             'dc': ''}

domainmap = {'constant': 'dc',
             'constant time': 'dc',
             'constant frequency response': 'dc',             # ????
             'time': 't',
             'laplace': 's',
             'fourier': 'f',
             'angular fourier': 'omega'}


def domain_kind_to_class(kind):

    if isinstance(kind, str) and kind[0] == 'n':
        return classmap['n']

    try:
        return classmap[kind]
    except:
        return PhasorDomainExpression


def domain_kind_quantity_to_class(kind, quantity='undefined'):

    cls = domain_kind_to_class(kind)
    # TODO: There must be a better way than creating a dummy object.
    return cls(0)._class_by_quantity(quantity)


def domain_quantity_to_class(domain, quantity='undefined'):

    kind = domainmap[domain]
    return domain_kind_quantity_to_class(kind, quantity)


def domain_kind_to_symbol(kind, name):

    try:
        return name + symbolmap[kind]
    except:
        return name + '(%s)' % kind


def domain_to_domain_kind(domain):

    try:
        return domainmap[domain]
    except:
        return None


def expr_to_domain_kind(expr):

    try:
        return domainmap[expr.domain]
    except:
        return None
